import datetime
import pdb

import shopify
import pyactiveresource
import shopify_python_api_patcher
from shopify_python_api_patcher.utils import get_current_release

from em_tasks import logger
from em_tasks.store.store_util import StoreUtil
from em_tasks.utils.sku_parser import SkuParser
from em_tasks.utils.exchange_rate import ExchangeRate
from em_tasks.store.models import *


class OrderUtil():
  def __init__(self, store_name, app_token, release=None):
    self.store_name = store_name
    self.app_token = app_token
    if release:
      self.release = release
    else:
      self.release = get_current_release()

    shop_url = "{}.myshopify.com".format(store_name)
    self.shopify_credentials = {'domain': shop_url, 'version': self.release, 'token': self.app_token}

    session = shopify.Session(shop_url, self.release, self.app_token)
    shopify.ShopifyResource.activate_session(session)
    self.shop = shopify.Shop.current()
    self.order_decimal_fields = [
      'total', 'item_total', 'tax', 'shipping_charge', 'discount', 'shipping_cost',
      'product_cost', 'refund_amount'
    ]
    self.order_item_decimal_fields = [
      's_total', 's_price', 's_tax', 's_discount', 's_shipping_charge'
    ]

  def import_disputes(self, params={}):
    records = dict()
    for item in self.list_disputes(params):
      dispute = item['dispute']
      order = item['order']
      # source = 'OnlineStore'
      # if not order:
      #   continue
      # if order.tags:
      #   for tag in order.tags.split(','):
      #     tag = tag.strip()
      #     if tag.lower().find('subscription') != -1:
      #       source = 'Subscription'
      #       break
      # elif order.note_attributes:
      #   for note_attr in order.note_attributes:
      #     if note_attr.name.lower() == '_source':
      #       source = note_attr.value
      #       break

      records[dispute.id] = {
        's_id': dispute.id,
        's_order_id': dispute.order_id,
        's_order_number': order.name,
        'type': dispute.type,
        'currency': dispute.currency,
        'amount': round(float(dispute.amount), 2),
        'reason': dispute.reason,
        'network_reason_code': dispute.network_reason_code,
        'status': dispute.status,
        'evidence_due_by': dispute.evidence_due_by,
        'evidence_sent_on': dispute.evidence_sent_on,
        'finalized_on': dispute.finalized_on,
      }

    dispute_ids = list(records.keys())
    for dispute in Dispute.select().where(Dispute.s_id.in_(dispute_ids)).execute():
      if dispute.s_id not in records:
        continue

      record = records.pop(dispute.s_id)
      is_changed = False
      for k, v in record.items():
        dv = getattr(dispute, k)
        if k in ['evidence_sent_on', 'finalized_on']:
          if dv is None and v:
            is_changed = True
            break
        else:
          if dv != v:
            is_changed = True
            break
      if is_changed:
        for k, v in record.items():
          setattr(dispute, k, v)

        dispute.save()

    if records:
      Dispute.insert_many(list(records.values())).execute()

  def import_orders(self, params={}):
    orders = dict()
    batch_size = 250
    for order in self.list_orders(params):
      order_attrs = self.format_order(order)
      if not order_attrs:
        continue

      try:
        order_db = Order.get(Order.s_order_id == order.id)
        order_changed = False
        for k, v in order_attrs.items():
          if k in self.order_decimal_fields:
            vb = getattr(order_db, k)
            vf = round(float(v), 2)
            if vb and round(vb, 2) != vf:
              logger.debug('[OrderChanged] Key: %s, Value: %s, OldValue: %s', k, v, vb)
              order_changed = True
              break
          elif k in ['purchase_date', 'risky', 'risky_action', 'risky_message']:
            pass
          elif k in ['cancelled_at', 'refunded_at', 'closed_at'] and getattr(order_db, k):
            pass
          else:
            if getattr(order_db, k) != v:
              logger.debug('[OrderChangedNoneDecimal] Key: %s, Value: %s, OldValue: %s', k, v, getattr(order_db, k))
              order_changed = True
              break

        if order_changed:
          for k, v in order_attrs.items():
            setattr(order_db, k, v)

          order_db.save()
      except Order.DoesNotExist:
        Order.insert(order_attrs).execute()
      except Exception as e:
        logger.exception(e)

      orders[order.id] = order
      if len(orders) < batch_size:
        continue

      self.process_orders(orders)
      orders.clear()

    if orders:
      self.process_orders(orders)

  def process_orders(self, orders):
    order_items_db = dict()
    order_ids_found = dict()
    for order_item_db in OrderItem.select().where(OrderItem.s_order_id.in_(list(orders.keys()))).execute():
      if order_item_db.s_order_id not in orders:
        continue

      order_item_key = '{}-{}'.format(order_item_db.s_order_id, order_item_db.s_line_item_id)
      order_items_db[order_item_key] = order_item_db
      order_ids_found[order_item_db.s_order_id] = None

    product_ids = dict()
    for order in orders.values():
      for line_item in order.line_items:
        product_ids[line_item.product_id] = None
    product_ids_s = ','.join([str(pid) for pid in product_ids.keys()])
    products = dict()
    for product in shopify.Product.find(ids=product_ids_s, limit=250):
      products[product.id] = product

    order_items = dict()
    for oid, order in orders.items():
      if oid in order_ids_found:
        formatted_order_items = self.format_order_items(order)
      else:
        formatted_order_items = self.format_order_items(order, True)

      if not formatted_order_items:
        continue

      for order_item in formatted_order_items:
        if order_item['s_product_id'] in products:
          product = products[order_item['s_product_id']]
          url, image_url, options_text = self.get_product_info(product, order_item['s_variant_id'])
          if not order_item['s_url']:
            order_item['s_url'] = url
          if not order_item['s_image_url']:
            order_item['s_image_url'] = image_url
          if not order_item['s_options_text']:
            order_item['s_options_text'] = options_text

        order_item_key = '{}-{}'.format(order_item['s_order_id'], order_item['s_line_item_id'])
        order_items[order_item_key] = order_item

    to_create = []
    for order_item_key, order_item in order_items.items():
      if order_item_key in order_items_db:
        order_item_db = order_items_db[order_item_key]
        order_item_changed = False
        for k, v in order_item.items():
          if k in self.order_item_decimal_fields:
            vb = getattr(order_item_db, k)
            vf = round(float(v), 2)
            if vb and round(vb, 2) != vf:
              logger.debug('[OrderItemChanged] Key: %s, Value: %s, OldValue: %s', k, v, vb)
              order_item_changed = True
              break
          elif k == 's_fulfillment_order_id':
            # Don't touch fulfilled orders
            if order_item_db.s_fulfillment_id:
              continue

            old_v = getattr(order_item_db, k)
            fulfillment_order_id = v
            if not v:
              fulfillment_orders = shopify.FulfillmentOrders.find(order_id=order_item_db.s_order_id)
              if fulfillment_orders:
                # logger.debug('[FulfillmentOrdersFound] Order: %s', order_item_db.s_order_number)
                fulfillment_order_selected = None
                for fulfillment_order in fulfillment_orders:
                  for fli in fulfillment_order.line_items:
                    if fli.line_item_id == order_item_db.s_line_item_id:
                      fulfillment_order_selected = fulfillment_order
                      break

                  if fulfillment_order_selected:
                    break

                if fulfillment_order_selected:
                  fulfillment_order_id = fulfillment_order_selected.id
              else:
                logger.info('[NoFulfillmentOrder] Order: %s', order_item_db.s_order_number)

            if fulfillment_order_id and str(fulfillment_order_id) != old_v:
              logger.debug('[FulfillmentOrderIDChanged] Order: %s, New: %s, Old: %s', order_item_db.s_order_number, fulfillment_order_id, old_v)
              order_item[k] = fulfillment_order_id
              order_item_changed = True
              break
          elif k in ['s_fulfillment_id']:
            if not v:
              continue

            vb = getattr(order_item_db, k)
            if not vb and v:
              order_item_changed = True
              break

            if int(vb) != int(v):
              order_item_changed = True
              break
          elif k in ['s_carrier', 's_tracking']:
            # Don't sync our own tracking
            if v and (v == 'EM' or v.startswith('CP')):
              continue

            # Don't sync international tracking
            s_intl_forward = getattr(order_item_db, 's_intl_forward')
            if s_intl_forward:
              continue

            vb = getattr(order_item_db, k)
            if vb != v:
              order_item_changed = True
          elif k in ['vendor_id', 's_product_id', 's_variant_id', 's_purchase_date', 'status', 'created_at', 'updated_at', 'risky', 'risky_action', 'risky_message', 's_min_delivery_date', 's_max_delivery_date']:
            pass
          elif k in ['cancelled_at', 'cancel_reason', 'closed_at', 'refunds', 'refund_at', 'refund_user_id']:
            # Only update when event happens and local does not exist
            vb = getattr(order_item_db, k)
            if v and not vb:
              order_item_changed = True
              break
          elif k in ['s_name', 's_url', 's_image_url', 's_options_text'] and getattr(order_item_db, k):
            if not getattr(order_item_db, k):
              order_item_changed = True
              break
          else:
            if getattr(order_item_db, k) != v:
              logger.debug('[OrderItemChangedNoneDecimal] Order: %s, Key: %s, Value: %s, OldValue: %s', order_item_db.s_order_number, k, v, getattr(order_item_db, k))
              order_item_changed = True
              break

        if order_item['canceled'] and order_item_db.status != StoreUtil.ORDER_ITEM_STATUSES['CANCELED']:
          order_item_changed = True
          order_item_db.status = StoreUtil.ORDER_ITEM_STATUSES['CANCELED']
        elif order_item_db.status is None:
          order_item_changed = True
          order_item_db.status = StoreUtil.ORDER_ITEM_STATUSES['PENDING']

        s_fulfillment_order_id = order_item.get('s_fulfillment_order_id', None)
        if not s_fulfillment_order_id:
          order_item.pop('s_fulfillment_order_id')
        order_item.pop('status', None)
        s_carrier = order_item.get('s_carrier', None)
        s_tracking = order_item.get('s_tracking', None)
        if (s_carrier and s_carrier == 'EM') or (s_tracking and s_tracking.startswith('CP')):
          order_item['s_carrier'] = order_item['s_tracking'] = None
        if order_item_changed:
          for k, v in order_item.items():
            setattr(order_item_db, k, v)

          order_item_db.save()
      else:
        to_create.append(order_item)

    if to_create:
      OrderItem.insert_many(to_create).execute()

  @classmethod
  def update_order_events(cls, days=7):
    today = datetime.datetime.utcnow()
    from_date = (today - datetime.timedelta(days=days)).date()

    # Order Sales
    event_type = 'OrderSales'
    order_items = dict()
    for order_item in OrderItem.select().where(OrderItem.s_purchase_date >= from_date).execute():
      order_items[order_item.id] = order_item

    order_item_ids = list(order_items.keys())
    for order_event in OrderEvent.select().where((OrderEvent.order_item_id.in_(order_item_ids)) & (OrderEvent.event_type == event_type)).execute():
      order_items.pop(order_event.order_item_id, None)

    order_sales_events = []
    for order_item_id, order_item in order_items.items():
      data = {
        'amount': order_item.s_total,
        'discount': order_item.s_discount,
        'tax': order_item.s_tax,
        'item_price': order_item.s_price,
        'quantity': order_item.s_quantity,
        'shipping_charge': order_item.s_shipping_charge,
      }
      order_sales_events.append({
        'order_item_id': order_item_id,
        'event_type': event_type,
        'event_time': order_item.s_purchase_date,
        'event_data': json.dumps(data)
      })
    OrderEvent.insert_many(order_sales_events).execute()

    logger.info('[OrderEvent] {} updated!'.format(event_type))

    # Order Cancel
    event_type = 'OrderCancel'
    order_items = dict()
    for order_item in OrderItem.select().where(OrderItem.cancelled_at >= from_date).execute():
      order_items[order_item.id] = order_item

    order_item_ids = list(order_items.keys())
    for order_event in OrderEvent.select().where((OrderEvent.order_item_id.in_(order_item_ids)) & (OrderEvent.event_type == event_type)).execute():
      order_items.pop(order_event.order_item_id, None)

    order_cancel_events = []
    for order_item_id, order_item in order_items.items():
      data = {
        'amount': order_item.s_total,
        'tax': order_item.s_tax,
        'reason': order_item.cancel_reason
      }
      order_cancel_events.append({
        'order_item_id': order_item_id,
        'event_type': event_type,
        'event_time': order_item.cancelled_at,
        'event_data': json.dumps(data)
      })
    OrderEvent.insert_many(order_cancel_events).execute()

    logger.info('[OrderEvent] {} updated!'.format(event_type))

    # Order Refund
    event_type = 'OrderRefund'
    order_items = dict()
    for order_item in OrderItem.select().where((OrderItem.refund_at >= from_date) & (OrderItem.canceled == 0)).execute():
      order_items[order_item.id] = order_item

    order_item_ids = list(order_items.keys())
    for order_event in OrderEvent.select().where((OrderEvent.order_item_id.in_(order_item_ids)) & (OrderEvent.event_type == event_type)).execute():
      order_items.pop(order_event.order_item_id, None)

    order_refund_events = []
    for order_item_id, order_item in order_items.items():
      data = {
        'amount': order_item.refund_amount,
        'user_id': order_item.refund_user_id
      }
      order_refund_events.append({
        'order_item_id': order_item_id,
        'event_type': event_type,
        'event_time': order_item.refund_at,
        'event_data': json.dumps(data)
      })
    OrderEvent.insert_many(order_refund_events).execute()

    order_item_ids = []
    for order_event in OrderEvent.select().where(OrderEvent.event_type == 'OrderRefund'):
      order_item_ids.append(order_event.order_item_id)

    canceled_order_item_ids = []
    for order_item in OrderItem.select().where((OrderItem.id.in_(order_item_ids)) & (OrderItem.canceled == 1)).execute():
      canceled_order_item_ids.append(order_item.id)
    OrderEvent.delete().where((OrderEvent.event_type == 'OrderRefund') & (OrderEvent.order_item_id.in_(canceled_order_item_ids))).execute()

    logger.info('[OrderEvent] {} updated!'.format(event_type))

    # Order Purchased
    event_type = 'OrderPurchased'
    order_items = dict()
    for order_item in OrderItem.select().where(OrderItem.s_purchase_date >= from_date).execute():
      order_items[order_item.id] = order_item

    order_item_ids = list(order_items.keys())
    fulfill_orders = dict()
    for fulfill_order in FulfillOrder.select().where(FulfillOrder.order_item_id.in_(order_item_ids)).execute():
      fulfill_orders[fulfill_order.order_item_id] = fulfill_order

    order_purchase_events = dict()
    for order_item_id, order_item in order_items.items():
      cost = order_item.source_price or 0
      order_cost_recorded = False
      if order_item.source == 'EMP':
        order_cost_recorded = True
      if order_item_id in fulfill_orders and fulfill_orders[order_item_id].buy_cost is not None and fulfill_orders[order_item_id].buy_cost:
        cost = fulfill_orders[order_item_id].buy_cost
        order_cost_recorded = True

      order_purchase_events[order_item_id] = {
        'order_item_id': order_item_id,
        'event_type': event_type,
        'event_time': order_item.s_purchase_date,
        'event_data': json.dumps({'amount': cost, 'recorded': order_cost_recorded})
      }

    for order_event in OrderEvent.select().where((OrderEvent.order_item_id.in_(order_item_ids)) & (OrderEvent.event_type == event_type)).execute():
      order_purchase_event = order_purchase_events.pop(order_event.order_item_id)
      if order_event.event_data != order_purchase_event['event_data']:
        order_event.event_data = order_purchase_event['event_data']
        order_event.save()

    if order_purchase_events:
      order_purchase_events = list(order_purchase_events.values())
      OrderEvent.insert_many(order_purchase_events).execute()

    logger.info('[OrderEvent] {} updated!'.format(event_type))

    # Order Shipped
    event_type = 'OrderShipped'
    order_items = dict()
    for order_item in OrderItem.select().where(OrderItem.s_purchase_date >= from_date).execute():
      order_items[order_item.id] = order_item

    order_item_ids = list(order_items.keys())
    for order_event in OrderEvent.select().where((OrderEvent.order_item_id.in_(order_item_ids)) & (OrderEvent.event_type == event_type)).execute():
      order_items.pop(order_event.order_item_id, None)

    order_shipped_events = []
    order_item_ids = list(order_items.keys())
    for fulfill_order in FulfillOrder.select().where(FulfillOrder.order_item_id.in_(order_item_ids)).execute():
      if not fulfill_order.stc:
        continue

      if fulfill_order.order_item_id not in order_items:
        continue

      order_item = order_items[fulfill_order.order_item_id]
      event_time = order_item.s_purchase_date
      order_shipped_events.append({
        'order_item_id': fulfill_order.order_item_id,
        'event_type': event_type,
        'event_time': event_time,
        'event_data': json.dumps({'amount': 0})
      })
    OrderEvent.insert_many(order_shipped_events).execute()

    order_shipped_events = dict()
    shipments = dict()
    for ss in ShipstationShipment.select().where(ShipstationShipment.create_date >= from_date).execute():
      shipments[ss.order_number] = ss

    shipstation_order_numbers = list(shipments.keys())
    for so in ShipstationOrder.select().where(ShipstationOrder.order_number.in_(shipstation_order_numbers)).execute():
      if so.order_number not in shipments:
        continue

      order_item_ids.append(so.order_item_id)
      shipment = shipments[so.order_number]
      order_shipped_events[so.order_item_id] = {
        'order_item_id': so.order_item_id,
        'event_type': event_type,
        'event_time': shipment.create_date,
        'event_data': json.dumps({'amount': round(shipment.shipment_cost + shipment.insurance_cost, 2)})
      }

    order_item_ids = list(order_shipped_events.keys())
    for order_event in OrderEvent.select().where((OrderEvent.order_item_id.in_(order_item_ids)) & (OrderEvent.event_type == event_type)).execute():
      order_shipped_event = order_shipped_events.pop(order_event.order_item_id, None)
      if not order_shipped_event:
        continue

      if order_event.event_data != order_shipped_event['event_data']:
        order_event.event_data = order_shipped_event['event_data']
        order_event.event_time = order_shipped_event['event_time']
        order_event.save()

    if order_shipped_events:
      OrderEvent.insert_many(list(order_shipped_events.values())).execute()

    logger.info('[OrderEvent] {} updated!'.format(event_type))

  def cancel_risk_orders(self, params={}):
    for order in self.list_orders(params):
      if order.cancelled_at:
        logger.debug('[OrderAlreadyCanceled] %s', order.name)
        continue

      if order.financial_status in ['voided', 'refunded']:
        logger.debug('[OrderVoidedOrRefunded] %s', order.name)
        continue

      should_cancel_order = False
      order_risks = shopify.OrderRisk.find(order_id=order.id)
      for order_risk in order_risks:
        if order_risk.recommendation == 'cancel':
          should_cancel_order = True
          break

      if not should_cancel_order:
        continue

      order.cancel(reason='fraud', email=True)
      self.void_order(order)

  def capture_fulfillable_orders(self):
    orders = dict()
    for order in Order.select(Order).join(OrderItem, on=(Order.s_order_id == OrderItem.s_order_id)).where((Order.financial_status == 'authorized') & (OrderItem.fulfillable == 1)).execute():
      orders[order.s_order_id] = order

    order_ids = list(orders.keys())
    chunks = [order_ids[x:x + 50] for x in range(0, len(order_ids), 50)]
    for chunk in chunks:
      ids = ','.join([str(oid) for oid in chunk])
      for shopify_order in shopify.Order.find(ids=ids):
        if shopify_order.financial_status == 'authorized':
          payloads = []
          invalid_ts = []
          transactions = shopify_order.transactions()
          for transaction in transactions:
            if transaction.kind != 'authorization':
              invalid_ts.append(transaction)
              continue

            payloads.append({
              'kind': 'capture',
              'order_id': shopify_order.id,
              'authorization': transaction.authorization,
              'amount': transaction.amount
            })

          if len(invalid_ts) >= len(payloads):
            logger.info(
              '[OrderPaymentError] Order: %s, OrderNumber: %s, Amount: %s', shopify_order.id, shopify_order.name, shopify_order.total_price)
            continue

          for payload in sorted(payloads, key=lambda p: p['amount']):
            transaction = shopify.Transaction.create(payload)
            # succeed = shopify_order.capture()
            if transaction:
              logger.info('[OrderCaptured] Order: %s, OrderNumber: %s, Amount: %s', shopify_order.id, shopify_order.name, payload['amount'])
            else:
              logger.info('[OrderCaptureFailed] Order: %s, OrderNumber: %s, Error: %s', shopify_order.id, shopify_order.name, transaction.errors.full_messages())

  def capture_fulfilled_orders(self, days=7):
    today = datetime.datetime.utcnow()
    from_date = today - datetime.timedelta(days=days)

    s_order_ids = dict()
    for order_item in OrderItem.select(OrderItem).join(FulfillOrder, on=(OrderItem.id == FulfillOrder.order_item_id)).where(OrderItem.s_purchase_date > from_date).execute():
      s_order_ids[order_item.s_order_id] = None

    if not s_order_ids:
      return

    s_order_ids = list(s_order_ids.keys())
    orders = dict()
    for order in Order.select().where((Order.s_order_id.in_(s_order_ids)) & (Order.financial_status == 'authorized')).execute():
      orders[order.s_order_id] = order

    order_ids = list(orders.keys())
    chunks = [order_ids[x:x + 50] for x in range(0, len(order_ids), 50)]
    for chunk in chunks:
      ids = ','.join([str(oid) for oid in chunk])
      for shopify_order in shopify.Order.find(ids=ids, status='any', limit=250):
        if shopify_order.financial_status == 'authorized':
          payloads = []
          transactions = shopify_order.transactions()
          invalid_ts = []
          for transaction in transactions:
            if transaction.kind != 'authorization':
              invalid_ts.append(transaction)
              continue

            payloads.append({
              'kind': 'capture',
              'order_id': shopify_order.id,
              'authorization': transaction.authorization,
              'amount': transaction.amount
            })

          if len(invalid_ts) >= len(payloads):
            logger.info(
              '[OrderPaymentError] Order: %s, OrderNumber: %s, Amount: %s', shopify_order.id, shopify_order.name, shopify_order.total_price)
            continue

          for payload in sorted(payloads, key=lambda p: p['amount']):
            transaction = shopify.Transaction.create(payload)
            # succeed = shopify_order.capture()
            if transaction:
              logger.info('[OrderCaptured] Order: %s, OrderNumber: %s, Amount: %s', shopify_order.id, shopify_order.name, payload['amount'])
            else:
              logger.info('[OrderCaptureFailed] Order: %s, OrderNumber: %s, Error: %s', shopify_order.id, shopify_order.name, transaction.errors.full_messages())

  def void_order(self, order):
    transactions = order.transactions()
    for transaction in transactions:
      if transaction.kind != 'authorization':
        continue

      payload = {
        'kind': 'void',
        'order_id': order.id,
      }

      t = shopify.Transaction.create(payload)
      # succeed = shopify_order.capture()
      if t:
        logger.info('[OrderVoided] Order: %s, OrderNumber: %s, PurchaseDate: %s, CancelledDate: %s, Amount: %s', order.id, order.name, order.created_at, order.cancelled_at, transaction.amount)
      else:
        logger.info('[OrderVoidFailed] Order: %s, OrderNumber: %s, PurchaseDate: %s, Error: %s', order.id, order.name, order.created_at, t.errors.full_messages())

  def upload_trackings(self, days=14):
    orders = dict()
    for order_item in OrderItem.select(OrderItem, FulfillOrder).join(FulfillOrder, on=(OrderItem.id == FulfillOrder.order_item_id), attr='fulfill_order').where((FulfillOrder.stc == 1) & (FulfillOrder.buy_tracking != '') & (FulfillOrder.buy_tracking.is_null(False)) & (OrderItem.s_fulfillment_id.is_null(True))).execute():
      if order_item.s_order_id not in orders:
        orders[order_item.s_order_id] = []
      orders[order_item.s_order_id].append(order_item)

    for order_item in OrderItem.select().where((OrderItem.s_intl_forward == 1) & (OrderItem.s_intl_tracking.is_null(False)) & (OrderItem.s_tracking.is_null(True)) & (OrderItem.s_fulfillment_id.is_null(True))).execute():
      if order_item.s_order_id not in orders:
        orders[order_item.s_order_id] = []
      orders[order_item.s_order_id].append(order_item)

    for order_item in OrderItem.select().where((OrderItem.s_tracking.is_null(False)) & (OrderItem.s_fulfillment_id.is_null(True))).execute():
      if order_item.s_order_id not in orders:
        orders[order_item.s_order_id] = []
      orders[order_item.s_order_id].append(order_item)

    order_ids = list(orders.keys())
    for order in Order.select().where(Order.s_order_id.in_(order_ids)).execute():
      if order.fulfillment_status == 'fulfilled':
        orders.pop(order.s_order_id, None)
        continue

      for order_item in orders[order.s_order_id]:
        self.upload_tracking(order_item)

  def upload_tracking(self, order_item):
    if order_item.canceled:
      logger.debug('[OrderCancelled] %s', order_item.s_order_number)
      return

    if order_item.closed:
      logger.debug('[OrderClosed] %s', order_item.s_order_number)
      return

    if order_item.s_fulfillable_quantity == 0:
      logger.debug('[OrderFulfillableQuantityZero] %s', order_item.s_order_number)
      return

    if order_item.s_fulfillment_status == 'fulfilled':
      logger.debug('[OrderFulfilled] %s', order_item.s_order_number)
      return

    if order_item.s_fulfillment_status in ['not_eligible', 'null']:
      logger.debug('[OrderNotEligibleToFulfill] %s', order_item.s_order_number)
      return

    fulfillment_order_id = order_item.s_fulfillment_order_id
    fulfillment_line_item_id = None
    try:
      # Shopify python API bug
      fo_prefix = shopify.FulfillmentOrders._prefix_source
      shopify.FulfillmentOrders._prefix_source = ''

      fulfillment_order = shopify.FulfillmentOrders.find(str(fulfillment_order_id))
      for li in fulfillment_order.line_items:
        if li.line_item_id == order_item.s_line_item_id:
          fulfillment_line_item_id = li.id
          break

      shopify.FulfillmentOrders._prefix_source = fo_prefix
    except Exception as e:
      logger.exception(e)

    quantity = order_item.s_quantity
    carrier = tracking = None
    if order_item.s_carrier:
      carrier = order_item.s_carrier
    elif order_item.s_intl_forward and order_item.s_intl_tracking:
      carrier = 'EM'
    else:
      fulfill_order = getattr(order_item, 'fulfill_order', None)
      if fulfill_order:
        carrier = fulfill_order.buy_carrier
    if order_item.s_tracking:
      tracking = order_item.s_tracking
    elif order_item.s_intl_forward and order_item.s_intl_tracking:
      tracking = order_item.s_intl_tracking
    else:
      fulfill_order = getattr(order_item, 'fulfill_order', None)
      if fulfill_order:
        tracking = fulfill_order.buy_tracking
    if not carrier or not tracking:
      return

    fulfillment = shopify.Fulfillment()
    fulfillment.notify_customer = True
    fulfillment.tracking_info = {
      'company': carrier,
      'number': tracking
    }
    if carrier == 'PARCELSAPP':
      fulfillment.tracking_info['url'] = 'https://parcelsapp.com/en/tracking/{}'.format(tracking)
    elif carrier == 'EM':
      fulfillment.tracking_info['url'] = 'https://tracking.everymarket.com/tracking/{}'.format(tracking)

    fo = [{
      'fulfillment_order_id': fulfillment_order_id,
    }]
    if fulfillment_line_item_id:
      fo = [{
        'fulfillment_order_id': fulfillment_order_id,
        'fulfillment_order_line_items': [{
          'id': fulfillment_line_item_id,
          'quantity': quantity
        }]
      }]

    fulfillment.line_items_by_fulfillment_order = fo
    try:
      # Shopify python API bug
      f_prefix = shopify.Fulfillment._prefix_source
      shopify.Fulfillment._prefix_source = ''

      succeed = fulfillment.save()

      shopify.Fulfillment._prefix_source = f_prefix

      if succeed:
        order_item.s_fulfillment_id = fulfillment.id
        if not order_item.s_intl_forward:
          order_item.s_carrier = carrier
          order_item.s_tracking = tracking
        order_item.save()
        logger.info(
          '[OrderTrackingUploaded] OrderNumber: %s, FulfillmentID: %s, Carrier: %s, Tracking: %s',
          order_item.s_order_number, order_item.s_fulfillment_id, carrier, tracking)
      else:
        logger.info(
          '[OrderTrackingUploadFailed] OrderNumber: %s, OrderID: %s, Carrier: %s, Tracking: %s, Error: %s, Payload: %s',
          order_item.s_order_number, order_item.s_order_id, carrier, tracking,
          fulfillment.errors.full_messages(), fulfillment.to_json())
    except Exception as e:
      logger.exception(e)
      logger.debug(fulfillment.to_json())

  @classmethod
  def find_amz_order_offers(self, offer_service, days=7):
    sources = ['AMZ_{}'.format(marketplace) for marketplace in SkuParser.supported_marketplaces]
    today = datetime.datetime.utcnow()
    from_date = today - datetime.timedelta(days=days)
    order_items = OrderItem.select().where((OrderItem.risky == 0) & (OrderItem.canceled == 0) & (OrderItem.source.in_(sources)) & OrderItem.source_price.is_null() & (OrderItem.s_purchase_date > from_date)).execute()

    order_items_by_m = dict()
    for order_item in order_items:
      marketplace = order_item.source.split('_').pop().lower()
      if marketplace not in order_items_by_m:
        order_items_by_m[marketplace] = dict()

      order_items_by_m[marketplace][order_item.source_id] = order_item

    for marketplace, order_items in order_items_by_m.items():
      asins = list(order_items.keys())
      offers = offer_service.get_offers(marketplace.lower(), asins)
      if not offers:
        continue

      for asin, offer in offers.items():
        if asin not in order_items:
          continue

        if not offer:
          continue

        order_item = order_items[asin]

        source_price_in_usd = offer['price']
        currency = offer.get('currency', 'USD')
        if currency != 'USD':
          exchange_rate = ExchangeRate.get_exchange_rate('USD', currency)
          source_price_in_usd = offer['price'] / exchange_rate

        order_item.source_price = round(source_price_in_usd, 2)
        order_item.estimate_profit = round(order_item.s_price - source_price_in_usd, 2)
        if order_item.estimate_profit > 0:
          order_item.fulfillable = True
        order_item.save()
        logger.info(
          '[OrderOfferFind] Order: %s, Price: %s, Source: %s, SourceId: %s, SourcePrice: %s',
          order_item.s_order_number, order_item.s_price, order_item.source, order_item.source_id, order_item.source_price)

  def list_orders(self, params={}):
    days = 20
    today = datetime.datetime.utcnow()
    from_date = today - datetime.timedelta(days=days)
    from_date_s = from_date.replace(microsecond=0).isoformat()
    page_size = 250
    dft_params = {'limit': page_size, 'processed_at_min': from_date_s, 'status': 'any'}
    # dft_params = {'limit': page_size, 'updated_at_min': from_date_s, 'status': 'any'}
    filters = dict(dft_params)
    for k, v in params.items():
      if k in filters:
        filters.pop(k)

      if v:
        filters[k] = v

    logger.debug('[OrderFilters] %s', filters)

    next_page_url = None
    has_more_orders = True
    with shopify.Session.temp(**self.shopify_credentials):
      while has_more_orders:
        if next_page_url:
          orders = shopify.Order.find(from_=next_page_url)
        else:
          orders = shopify.Order.find(**filters)

        if not orders:
          has_more_orders = False
          break

        print('[OrdersNextPage] {}'.format(orders.next_page_url))
        if len(orders) < page_size:
          has_more_orders = False
        else:
          if next_page_url == orders.next_page_url:
            break

          next_page_url = orders.next_page_url

        for order in orders:
          yield order

  def list_disputes(self, params={}):
    next_page_url = None
    has_more_disputes = True
    with shopify.Session.temp(**self.shopify_credentials):
      while has_more_disputes:
        if next_page_url:
          disputes = shopify.Disputes.find(from_=next_page_url)
        else:
          if params:
            disputes = shopify.Disputes.find(**params)
          else:
            disputes = shopify.Disputes.find()

        if not disputes:
          has_more_disputes = False
          break

        print('[DisputesNextPage] {}'.format(disputes.next_page_url))
        if next_page_url == disputes.next_page_url:
          break

        next_page_url = disputes.next_page_url

        order_ids = list(set([dispute.order_id for dispute in disputes]))
        order_ids_s = ','.join([str(order_id) for order_id in order_ids])
        orders = dict()
        for order in shopify.Order.find(ids=order_ids_s, limit=len(order_ids), status='any'):
          orders[order.id] = order

        for dispute in disputes:
          yield {'dispute': dispute, 'order': orders.get(dispute.order_id, None)}

        if not next_page_url:
          break

  def format_order(self, shopify_order):
    if shopify_order.test:
      return

    shipping_charge = round(sum([float(shipping_line.discounted_price) for shipping_line in shopify_order.shipping_lines]), 2)
    order = {
      's_order_id': shopify_order.id,
      's_order_number': shopify_order.name,
      'purchase_date': shopify_order.processed_at,
      'total': round(float(shopify_order.total_price), 2),
      'item_total': shopify_order.total_line_items_price,
      'tax': shopify_order.total_tax,
      'shipping_charge': shipping_charge,
      'discount': shopify_order.total_discounts,
      'cancelled_at': shopify_order.cancelled_at,
      'cancel_reason': shopify_order.cancel_reason,
      'closed_at': shopify_order.closed_at,
      'financial_status': shopify_order.financial_status,
      'fulfillment_status': shopify_order.fulfillment_status,
      'note': shopify_order.note,
      'order_status_url': shopify_order.order_status_url
    }
    if not shopify_order.refunds:
      return order

    refund_amount = 0
    for refund in shopify_order.refunds:
      refund_amount += sum([float(transaction.amount) for transaction in refund.transactions])
    refund = shopify_order.refunds.pop()
    order['refunded_at'] = refund.processed_at
    order['refund_amount'] = refund_amount

    return order

  def format_order_items(self, shopify_order, verbose=False):
    if shopify_order.test:
      return

    order_items = []

    sa = shopify_order.shipping_address
    shipping_address = {
      'first_name': sa.first_name,
      'last_name': sa.last_name,
      'address1': sa.address1,
      'address2': sa.address2,
      'company': sa.company,
      'city': sa.city,
      'state': sa.province_code,
      'zipcode': sa.zip,
      'country': sa.country_code,
      'phone': sa.phone,
    }
    ba = shopify_order.billing_address
    billing_address = {
      'first_name': ba.first_name,
      'last_name': ba.last_name,
      'address1': ba.address1,
      'address2': ba.address2,
      'company': ba.company,
      'city': ba.city,
      'state': ba.province_code,
      'zipcode': ba.zip,
      'country': ba.country_code,
      'phone': ba.phone,
    }

    shippings = dict()
    for shipment in order['shipments']:
      selected_shipping_rate = shipment.get('selected_shipping_rate', None)
      if not selected_shipping_rate:
        for sr in shipment.get('shipping_rates', []):
          if sr.get('selected', False):
            selected_shipping_rate = sr
            break

      if not selected_shipping_rate:
        shipment.get('shipping_rates', [])[0]

      shipping_rate = {
        'name': selected_shipping_rate['name'],
        'cost': round(float(selected_shipping_rate['cost']), 2),
      }
      for manifest in shipment.get('manifest', []):
        variant_id = manifest.get('variant_id', None)
        if not variant_id:
          continue

        shippings[variant_id] = {
          'variant_id': variant_id,
          'quantity': manifest.get('quantity', 1),
          'shipping_method': shipping_rate['name'],
          'shipping_charge': shipping_rate['cost'],
          'shipping_state': shipment['state']
        }

    for line_item in order['line_items']:
      vendor_id = 1
      source = None
      source_id = None
      product_id = line_item.product_id
      store_product = StoreProduct.get_or_none(StoreProduct.s_product_id == product_id)
      if store_product:
        vendor_id = store_product.store_seller_id
        source = store_product.source
        source_id = store_product.source_product_id
      else:
        sku = line_item.sku
        parsed_vendor_id, source, source_id = SkuParser.parse(sku)
        if parsed_vendor_id:
          vendor_id = parsed_vendor_id

      product = None
      url = None
      image_url = None
      options_text = None
      # if product_id and verbose:
      #   try:
      #     product = shopify.Product.find(product_id)
      #     url, image_url, options_text = self.get_product_info(product)
      #   except pyactiveresource.connection.ResourceNotFound:
      #     pass
      fulfillment_selected = None
      for fulfillment in fulfillments:
        for fli in fulfillment.line_items:
          if fli.id == line_item.id:
            fulfillment_selected = fulfillment
            break

          try:
            if fli.fulfillment_line_item_id == line_item.id:
              fulfillment_selected = fulfillment
              break
          except:
            pass

        if fulfillment_selected:
          break

      price = round(float(line_item.price), 2)
      tax = round(sum([float(tax_line.price) for tax_line in line_item.tax_lines]), 2)
      discount = round(float(line_item.total_discount), 2)
      removed = line_item.fulfillable_quantity == 0 and line_item.fulfillment_status in ['null', 'partial', 'not_eligible']

      total = round(line_item['total'], 2)
      price = round(line_item['price'], 2)
      tax_adjustments = [adjustment for adjustment in line_item.get('adjustments', []) if adjustment['source_type'] == 'Spree::TaxRate']
      tax = round(sum([float(adjustment['amount']) for adjustment in tax_adjustments]), 2)
      quantity = int(line_item['quantity'])
      shipping = shippings.get(line_item['variant_id'], None)
      fulfillable_quantity = shipping['quantity'] if shipping else 0
      fulfillment_status = shipping['state'] if shipping else None

      order_item = {
        'email': order['email'],
        # Vendor
        # 'vendor_id': vendor_id,
        's_order_id': order['id'],
        's_order_number': order['number'],
        's_purchase_date': order['completed_at'],
        's_total': total,
        's_price': price,
        's_tax': tax,
        's_discount': discount,
        's_quantity': quantity,
        's_fulfillable_quantity': fulfillable_quantity,
        's_fulfillment_status': fulfillment_status,
        # Closed
        # 'closed': closed,
        # 'closed_at': shopify_order.closed_at,
        's_line_item_id': line_item['id'],
        's_product_id': line_item['variant']['product_id'],
        's_variant_id': line_item['variant_id'],
        's_name': line_item['variant']['name'],
        's_sku': line_item['variant']['sku'],
        's_url': '{}/products/{}'.format(self.shop.domain),
        's_image_url': image_url,
        's_options_text': options_text,
        's_shipping_method': shipping_line.code,
        's_shipping_charge': round(float(shipping_line.discounted_price), 2),
        'source': source,
        'source_id': source_id,
        'canceled': canceled,
        'cancelled_at': shopify_order.cancelled_at,
        'cancel_reason': shopify_order.cancel_reason,
        's_removed': removed,
        'status': StoreUtil.ORDER_ITEM_STATUSES['CANCELED'] if canceled else StoreUtil.ORDER_ITEM_STATUSES['PENDING'],
        'note': shopify_order.note,
        'created_at': shopify_order.created_at,
        'updated_at': shopify_order.updated_at
      }

      order_item.update(shipping_address)
      order_item['billing_address'] = json.dumps(billing_address)

      if fulfillment_selected:
        order_item['s_fulfillment_id'] = fulfillment_selected.id
        order_item['s_carrier'] = fulfillment_selected.tracking_company
        order_item['s_tracking'] = fulfillment_selected.tracking_number
        # order_item['s_tracking_url'] = fulfillment_selected.tracking_url
        # order_item['s_shipping_cost']
        order_item['status'] = fulfillment_selected.status
      else:
        order_item['status'] = shopify_order.financial_status

      if fulfillment_orders:
        fulfillment_order_selected = None
        for fulfillment_order in fulfillment_orders:
          for fli in fulfillment_order.line_items:
            if fli.line_item_id == line_item.id:
              fulfillment_order_selected = fulfillment_order
              break

          if fulfillment_order_selected:
            break

        if fulfillment_order_selected:
          order_item['s_fulfillment_order_id'] = fulfillment_order_selected.id
          delivery_method = fulfillment_order_selected.delivery_method
          if delivery_method.min_delivery_date_time:
            order_item['s_min_delivery_date'] = delivery_method.min_delivery_date_time

          if delivery_method.max_delivery_date_time:
            order_item['s_max_delivery_date'] = delivery_method.max_delivery_date_time
        else:
          order_item['s_fulfillment_order_id'] = None
          order_item['s_min_delivery_date'] = None
          order_item['s_max_delivery_date'] = None
      else:
        order_item['s_fulfillment_order_id'] = None
        order_item['s_min_delivery_date'] = None
        order_item['s_max_delivery_date'] = None

      if order_risks:
        order_risk_selected = None
        for order_risk in order_risks:
          if order_risk.recommendation != 'accept':
            order_risk_selected = order_risk
            break

        if not order_risk_selected and order_risks:
          order_risk_selected = order_risks[0]

        if order_risk_selected:
          order_item['risky'] = order_risk_selected.recommendation == 'cancel'
          order_item['risky_action'] = order_risk_selected.recommendation
          order_item['risky_message'] = order_risk_selected.message
        else:
          order_item['risky'] = False
          order_item['risky_action'] = None
          order_item['risky_message'] = None
      else:
        order_item['risky'] = False
        order_item['risky_action'] = None
        order_item['risky_message'] = None

      refund_selected = None
      refund_line_item_selected = None
      for refund in refunds:
        for refund_line_item in refund.refund_line_items:
          if refund_line_item.line_item_id == line_item.id:
            refund_line_item_selected = refund_line_item
            break

        if refund_line_item_selected:
          refund_selected = refund
          break

      if refund_line_item_selected:
        order_item['refund_qty'] = refund_line_item_selected.quantity
        order_item['refund_amount'] = round(float(refund_line_item_selected.subtotal) + float(refund_line_item_selected.total_tax), 2)
        order_item['refund_note'] = refund_selected.note
        order_item['refund_at'] = refund_selected.processed_at
        order_item['refund_user_id'] = refund_selected.user_id
        order_item['refunds'] = json.dumps([refund_selected.to_dict()])
      else:
        if financial_status.find('refund') != -1 and len(refunds) > 0:
          refund_selected = None
          amount = None
          for refund in refunds:
            for order_adjustment in refund.order_adjustments:
              # 4.99 is shipping charge
              adjustment_amount = round(float(order_adjustment.amount), 2)
              if abs(adjustment_amount - order_item['s_total']) <= 4.99:
                refund_selected = refund
                amount = adjustment_amount
                break
          # if financial_status == 'refunded':
          #   order_item['refund_qty'] = order_item['s_quantity']
          #   order_item['refund_amount'] = order_item['s_total']
          #   order_item['refund_note'] = refund_selected.note
          #   order_item['refund_at'] = refund_selected.processed_at
          #   order_item['refund_user_id'] = refund_selected.user_id
          #   order_item['refunds'] = json.dumps([refund_selected.to_dict()])
          # else:
          #   amount = None
          #   for refund in refunds:
          #     for order_adjustment in refund.order_adjustments:
          #       # 4.99 is shipping charge
          #       adjustment_amount = round(float(order_adjustment.amount), 2)
          #       if abs(adjustment_amount - order_item['s_total']) <= 4.99:
          #         refund_selected = refund
          #         amount = adjustment_amount
          #         break

          if refund_selected:
            order_item['refund_qty'] = order_item['s_quantity']
            order_item['refund_amount'] = amount or order_item['s_total']
            order_item['refund_note'] = refund_selected.note
            order_item['refund_at'] = refund_selected.processed_at
            order_item['refund_user_id'] = refund_selected.user_id
            order_item['refunds'] = json.dumps([refund_selected.to_dict()])
        else:
          if shopify_order.financial_status.find('refund') != -1 or len(refunds) > 0:
            logger.warning(
              '[OrderWithRefund] OrderNumber: %s, FinancialStatus: %s, Refunds: %s',
              shopify_order.name, shopify_order.financial_status, [refund.to_dict() for refund in shopify_order.refunds])
            # pdb.set_trace()
          order_item['refund_qty'] = 0
          order_item['refund_amount'] = 0
          order_item['refund_note'] = None
          order_item['refund_at'] = None
          order_item['refund_user_id'] = None
          order_item['refunds'] = None

      order_items.append(order_item)

    return order_items

  def get_product_info(self, product, variant_id=None):
    url = None
    if variant_id:
      url = 'https://{}/products/{}?variant_id={}'.format(self.shop.domain, product.handle, variant_id)
    else:
      url = 'https://{}/products/{}'.format(self.shop.domain, product.handle)

    image_url = None
    for image in product.images:
      if variant_id in image.variant_ids:
        image_url = image.src
        break
    if not image_url and product.images:
      image_url = product.images[0].src

    variant = product.variants[0]
    options_text = ''
    for v in product.variants:
      if v.id == variant_id:
        variant = v
        break
    if variant:
      if variant.option1 != 'Default Title':
        option_texts = []
        for option in product.options:
          option_text = '{}:{}'.format(option.name, getattr(variant, 'option{}'.format(option.position)))
          option_texts.append(option_text)

        options_text = ';'.join(option_texts)

    return (url, image_url, options_text)
