import os
import datetime
from urllib3.util import Retry

import requests
from requests.adapters import HTTPAdapter

from em_tasks import logger


class SpreeApi(object):
  def __init__(self, endpoint, api_key, api_version='v1'):
    self.endpoint = endpoint
    self.api_key = api_key
    self.api_version = api_version

    retries = Retry(
      total=7,
      backoff_factor=0.1,
      status_forcelist=[429, 500, 502, 503, 504],
    )
    self.session = requests.Session()
    self.session.mount(self.endpoint, HTTPAdapter(max_retries=retries))

  def list_orders(self, params={}):
    endpoint_url = '{}/api/{}/orders'.format(self.endpoint, self.api_version)
    payload = {
      'token': self.api_key,
      'q[completed_at_not_null]': 1,
      'q[s]': 'completed_at:asc',
      'page': 1,
      'per_page': 50
    }
    payload.update(params)

    # params = '&'.join([k if v is None else '{}={}'.format(k, v) for k, v in payload.items()])
    response = self.session.get(endpoint_url, params=payload)
    logger.debug('[Request] %s', response.request.url)

    return response.json()

  def get_orders(self, order_ids):
    endpoint_url = '{}/api/{}/orders'.format(self.endpoint, self.api_version)
    payload = {
      'token': self.api_key,
      'q[id_in][]': order_ids,
      'q[s]': 'completed_at:asc',
      'page': 1,
      'per_page': len(order_ids)
    }

    response = self.session.get(endpoint_url, params=payload)
    logger.debug('[Request] %s', response.request.url)

    return response.json()

  def approve_order(self, order_number):
    endpoint_url = '{}/api/{}/orders/{}/approve.json'.format(self.endpoint, self.api_version, order_number)
    payload = {'token': self.api_key}
    response = self.session.put(endpoint_url, params=payload)
    logger.debug('[Request] %s', response.request.url)

    return response.json()

  def get_order_refunds(self, order_number):
    endpoint_url = '{}/api/{}/orders/{}/refunds.json'.format(self.endpoint, self.api_version, order_number)
    payload = {'token': self.api_key}
    response = self.session.get(endpoint_url, params=payload)

    return response.json()

  def add_shipment_tracking(self, shipment_id, carrier, tracking, cost=None):
    endpoint_url = '{}/api/{}/shipments/{}/ship?token={}'.format(self.endpoint, self.api_version, shipment_id, self.api_key)
    payload = {'shipment': {'tracking': tracking, 'carrier': carrier}}
    if cost:
      payload['shipment']['actual_cost'] = cost

    response = self.session.put(endpoint_url, json=payload)

    return response.json()

  def update_stock(self, stock_item_id, qty, stock_location_id=1, force=True, backorderable=False):
    endpoint_url = '{}/api/{}/stock_locations/{}/stock_items/{}?token={}'.format(self.endpoint, self.api_version, stock_location_id, stock_item_id, self.api_key)
    payload = {'stock_item': {'count_on_hand': qty, 'force': force, 'backorderable': backorderable}}

    return self.session.put(endpoint_url, json=payload)

  def update_variant(self, product_id, variant_id, **params):
    endpoint_url = '{}/api/{}/products/{}/variants/{}?token={}'.format(self.endpoint, self.api_version, product_id, variant_id, self.api_key)
    payload = {'variant': params}

    return self.session.put(endpoint_url, json=payload)

  def download_inventory(self, source, output_path):
    if not output_path:
      return

    output_dir = os.path.dirname(output_path)
    if not os.path.isdir(output_dir):
      os.makedirs(output_dir)

    if os.path.isfile(output_path):
      os.remove(output_path)

    endpoint_url = '{}/api/{}/inventory_reports/download'.format(self.endpoint, self.api_version)
    payload = {'token': self.api_key, 'source': source}

    with self.session.get(endpoint_url, params=payload, stream=True) as r:
      r.raise_for_status()
      with open(output_path, 'wb+') as fh:
        for chunk in r.iter_content(chunk_size=8192):
          fh.write(chunk)

  def get_manual_products(self):
    endpoint_url = '{}/api/{}/inventory_reports/download_manual_products'.format(self.endpoint, self.api_version)
    payload = {'token': self.api_key}

    response = self.session.get(endpoint_url, params=payload)

    return response.json()

  def set_offers(self, product_offers):
    endpoint_url = '{}/api/{}/products/set_offers?token={}'.format(self.endpoint, self.api_version, self.api_key)
    payload = {
      'offers': product_offers,
    }

    response = self.session.post(endpoint_url, json=payload)

    return response.json()

  def add_fulfill_orders(self, fulfill_orders):
    endpoint_url = '{}/api/{}/fulfill_orders?token={}'.format(self.endpoint, self.api_version, self.api_key)
    payload = {'fulfill_orders': fulfill_orders}

    response = self.session.post(endpoint_url, json=payload)

    return response.json()

  def import_products(self, products, stock_location_id, shipping_category_id, vendor_id, merchant_id, min_shipping_days=7, taxonomy_name='Categories', tax_category_id=1):
    endpoint_url = '{}/api/{}/products/import?token={}'.format(self.endpoint, self.api_version, self.api_key)
    payload = {
      'products': products,
      'min_shipping_days': min_shipping_days,
      'taxonomy_name': taxonomy_name,
      'merchant_id': merchant_id,
      'vendor_id': vendor_id,
      'tax_category_id': tax_category_id,
      'stock_location_id': stock_location_id,
      'shipping_category_id': shipping_category_id
    }

    response = self.session.post(endpoint_url, json=payload)

    return response.json()

  def delete_products(self, product_ids):
    endpoint_url = '{}/api/{}/products/batch_delete?token={}'.format(self.endpoint, self.api_version, self.api_key)
    payload = {'product_ids': ','.join(product_ids)}

    response = self.session.post(endpoint_url, json=payload)

    return response.json()

  def get_shop(self):
    endpoint_url = '{}/api/v1/stores'.format(self.endpoint, self.api_version)
    response = self.session.get(endpoint_url, params={'token': self.api_key})

    resp = response.json()
    if 'stores' not in resp or not resp['stores']:
      return

    shop = None
    for store in resp['stores']:
      if store['default']:
        shop = store
        break

    if shop is None:
      shop = resp['stores'][0]

    return shop
