import datetime
import json
from urllib.parse import urlparse
import pdb

from em_tasks.spree.api import SpreeApi
from em_tasks.store.store_util import StoreUtil
from em_tasks.utils.sku_parser import SkuParser
from em_tasks.store.models import *

from em_tasks import logger


class OrderUtil():
  vendors = {
    24: {'code': 'em', 'vendor_id': 1},
    25: {'code': 'amazing-korea-1', 'vendor_id': 1},
    26: {'code': 'cairomy', 'vendor_id': 3},
    27: {'code': 'ilriposojewelry', 'vendor_id': None},
    28: {'code': 'ocdebiz153', 'vendor_id': 10},
    29: {'code': 'japan_devotion924', 'vendor_id': 4},
    30: {'code': 'korea-aroamas', 'vendor_id': 5},
    31: {'code': 'ocusbiz153', 'vendor_id': 6},
    32: {'code': 'japan-vitality', 'vendor_id': 7},
    33: {'code': 'em-uk', 'vendor_id': 8},
    34: {'code': 'em-tr', 'vendor_id': 9},
    35: {'code': 'elim-em', 'vendor_id': 10},
    36: {'code': 'em-cn', 'vendor_id': 11},
    37: {'code': 'em-tw', 'vendor_id': 12},
    38: {'code': 'amazing-korea', 'vendor_id': 2},
  }
  order_decimal_fields = [
    'total', 'item_total', 'tax', 'shipping_charge', 'discount', 'shipping_cost',
    'product_cost', 'refund_amount'
  ]
  order_item_decimal_fields = [
    's_total', 's_price', 's_tax', 's_discount', 's_shipping_charge'
  ]

  def __init__(self, endpoint, api_key, api_version='v1'):
    self.spree_api = SpreeApi(endpoint, api_key, api_version)
    self.shop = self.spree_api.get_shop()
    url = self.shop['url'] if self.shop['url'].startswith('http') else 'https://{}'.format(self.shop['url'])
    parsed = urlparse(self.shop['url'])
    self.shop_domain = parsed.hostname if parsed.hostname else 'everymarket.com'

  def import_orders(self, params={}):
    orders = dict()

    batch_size = 250
    for order in self.list_orders(params):
      orders[order['id']] = order
      if len(orders) < batch_size:
        continue

      self.process_orders(orders)
      orders.clear()

    if orders:
      self.process_orders(orders)
      orders.clear()

  def process_orders(self, orders):
    order_ids = list(orders.keys())

    orders_db = dict()
    for order in Order.select().where(Order.s_order_id.in_(order_ids)).execute():
      orders_db[order.s_order_id] = order

    orders_to_create = []
    for s_order_id, order in orders.items():
      order_attrs = self.format_order(order)
      if not order_attrs:
        continue

      if order['id'] not in orders_db:
        orders_to_create.append(order_attrs)
        continue

      order_changed = False
      order_db = orders_db[order['id']]
      for k, v in order_attrs.items():
        if k in self.order_decimal_fields:
          vb = getattr(order_db, k)
          vf = round(float(v), 2)
          if vb and round(vb, 2) != vf:
            logger.debug('[OrderChanged] Key: %s, Value: %s, OldValue: %s', k, v, vb)
            order_changed = True
            break
        elif k in ['purchase_date', 'risky', 'risky_action', 'risky_message']:
          pass
        elif k in ['cancelled_at', 'refunded_at', 'closed_at'] and getattr(order_db, k):
          pass
        else:
          if getattr(order_db, k) != v:
            logger.debug('[OrderChangedNoneDecimal] Key: %s, Value: %s, OldValue: %s', k, v, getattr(order_db, k))
            order_changed = True
            break

      if order_changed:
        for k, v in order_attrs.items():
          setattr(order_db, k, v)

        order_db.save()

    if orders_to_create:
      Order.insert_many(orders_to_create).execute()
      orders_to_create = []

    # Update order items
    order_items_db = dict()
    for order_item_db in OrderItem.select().where(OrderItem.s_order_id.in_(order_ids)).execute():
      if order_item_db.s_order_id not in orders:
        continue

      order_item_key = '{}-{}'.format(order_item_db.s_order_id, order_item_db.s_line_item_id)
      order_items_db[order_item_key] = order_item_db

    order_items = dict()
    for oid, order in orders.items():
      formatted_order_items = self.format_order_items(order)
      if not formatted_order_items:
        continue

      for order_item in formatted_order_items:
        order_item_key = '{}-{}'.format(order_item['s_order_id'], order_item['s_line_item_id'])
        order_items[order_item_key] = order_item

    to_create = []
    for order_item_key, order_item in order_items.items():
      if order_item_key in order_items_db:
        order_item_db = order_items_db[order_item_key]
        order_item_changed = False
        for k, v in order_item.items():
          if k in self.order_item_decimal_fields:
            vb = getattr(order_item_db, k)
            vf = round(float(v), 2)
            if vb and round(vb, 2) != vf:
              logger.debug('[OrderItemChanged] Key: %s, Value: %s, OldValue: %s', k, v, vb)
              order_item_changed = True
              break
          elif k in ['s_fulfillment_order_id', 's_fulfillment_id']:
            vb = getattr(order_item_db, k)
            if str(vb) != v:
              order_item_changed = True
          elif k in ['s_carrier', 's_tracking']:
            # Don't sync our own tracking
            if v and (v == 'EM' or v.startswith('CP')):
              continue

            # Don't sync international tracking
            s_intl_forward = getattr(order_item_db, 's_intl_forward')
            if s_intl_forward:
              continue

            vb = getattr(order_item_db, k)
            if vb != v:
              order_item_changed = True
          elif k in ['vendor_id', 's_product_id', 's_variant_id', 's_purchase_date', 'status', 'created_at', 'updated_at', 'risky', 'risky_action', 'risky_message', 's_min_delivery_date', 's_max_delivery_date']:
            pass
          elif k in ['cancelled_at', 'cancel_reason', 'closed_at', 'refunds', 'refund_at', 'refund_user_id']:
            # Only update when event happens and local does not exist
            vb = getattr(order_item_db, k)
            if v and not vb:
              order_item_changed = True
              break
          elif k in ['s_name', 's_url', 's_image_url', 's_options_text'] and getattr(order_item_db, k):
            if not getattr(order_item_db, k):
              order_item_changed = True
              break
          else:
            if getattr(order_item_db, k) != v:
              logger.debug('[OrderItemChangedNoneDecimal] Order: %s, Key: %s, Value: %s, OldValue: %s', order_item_db.s_order_number, k, v, getattr(order_item_db, k))
              order_item_changed = True
              break

        if order_item['canceled'] and order_item_db.status != StoreUtil.ORDER_ITEM_STATUSES['CANCELED']:
          order_item_changed = True
          order_item_db.status = StoreUtil.ORDER_ITEM_STATUSES['CANCELED']
        elif order_item_db.status is None:
          order_item_changed = True
          order_item_db.status = StoreUtil.ORDER_ITEM_STATUSES['PENDING']

        order_item.pop('status', None)
        s_carrier = order_item.get('s_carrier', None)
        s_tracking = order_item.get('s_tracking', None)
        if (s_carrier and s_carrier == 'EM') or (s_tracking and s_tracking.startswith('CP')):
          order_item['s_carrier'] = order_item['s_tracking'] = None
        if order_item_changed:
          for k, v in order_item.items():
            setattr(order_item_db, k, v)

          order_item_db.save()
      else:
        to_create.append(order_item)

    if to_create:
      OrderItem.insert_many(to_create).execute()

  def format_order(self, order):
    promo_total = order.get('promo_total', 0)
    promo_total = round(float(promo_total), 2) if promo_total else 0

    s_order = {
      's_order_id': order['id'],
      's_order_number': order['number'],
      'purchase_date': order['completed_at'],
      'total': round(float(order['total']), 2),
      'item_total': round(float(order['item_total']), 2),
      'tax': round(float(order['tax_total']), 2),
      'shipping_charge': round(float(order['ship_total'])),
      'discount': promo_total,
      'cancelled_at': order.get('canceled_at', None),
      'cancel_reason': None,
      'closed_at': None,
      'financial_status': order['payment_state'],
      'fulfillment_status': order['shipment_state'],
      'note': order.get('special_instructions', None),
      'order_status_url': 'https://{}/orders/{}?token={}'.format(self.shop['url'], order['number'], order['token'])
    }
    refunds = order.get('refunds', None)
    if not refunds:
      return s_order

    refund_amount = 0
    for refund in refunds:
      refund_amount += refund['amount']

    s_order['refunded_at'] = refund['created_at']
    s_order['refund_amount'] = refund_amount

    return s_order

  def format_order_items(self, order):
    d = dict()
    order_items = []

    sa = order['ship_address']
    shipping_address = {
      'first_name': sa['firstname'],
      'last_name': sa['lastname'],
      'address1': sa['address1'],
      'address2': sa.get('address2', None),
      'company': sa.get('company', None),
      'city': sa.get('city', None),
      'state': sa.get('state_text', sa.get('state', d).get('abbr', None)),
      'zipcode': sa.get('zipcode', None),
      'country': sa.get('country', d).get('iso', None),
      'phone': sa.get('phone', None),
    }
    ba = order['bill_address']
    billing_address = {
      'first_name': ba['firstname'],
      'last_name': ba['lastname'],
      'address1': ba['address1'],
      'address2': ba.get('address2', None),
      'company': ba.get('company', None),
      'city': ba.get('city', None),
      'state': ba.get('state_text', ba.get('state', d).get('abbr', None)),
      'zipcode': ba.get('zipcode', None),
      'country': ba.get('country', d).get('iso', None),
      'phone': ba.get('phone', None),
    }

    cancelled = bool(order['canceled_at'])
    refunds = order.get('refunds', None)

    shippings = dict()
    for shipment in order['shipments']:
      selected_shipping_rate = shipment.get('selected_shipping_rate', None)
      if not selected_shipping_rate:
        for sr in shipment.get('shipping_rates', []):
          if sr.get('selected', False):
            selected_shipping_rate = sr
            break

      if not selected_shipping_rate:
        shipment.get('shipping_rates', [])[0]

      shipping_rate = {
        'name': selected_shipping_rate['name'],
        'cost': round(float(shipment['actual_cost']), 2),
      }
      for manifest in shipment.get('manifest', []):
        variant_id = manifest.get('variant_id', None)
        if not variant_id:
          continue

        shippings[variant_id] = {
          'variant_id': variant_id,
          'quantity': manifest.get('quantity', 1),
          'shipment_id': shipment['number'],
          'shipment_number': shipment['number'],
          'shipping_method': shipping_rate['name'],
          'shipping_charge': shipping_rate['cost'],
          'shipping_state': shipment['state'],
          'shipped_at': shipment['shipped_at'],
          'carrier': shipment.get('carrier', None),
          'tracking': shipment.get('tracking', None)
        }

    for line_item in order['line_items']:
      variant = line_item['variant']
      sku = variant['sku']

      vendor_id = 1
      if line_item['vendor_id'] in self.vendors:
        vendor_id = self.vendors[line_item['vendor_id']]['vendor_id']

      source = source_id = None
      _, source, source_id = SkuParser.parse(sku)

      total = round(float(line_item['total']), 2)
      price = round(float(line_item['price']), 2)
      tax_adjustments = [adjustment for adjustment in line_item.get('adjustments', []) if adjustment['source_type'] == 'Spree::TaxRate']
      tax = round(sum([float(adjustment['amount']) for adjustment in tax_adjustments]), 2)
      discount_adjustments = [adjustment for adjustment in line_item.get('adjustments', []) if adjustment['source_type'] == 'Spree::PromotionAction']
      discount = round(-1 * sum([float(adjustment['amount']) for adjustment in discount_adjustments]), 2)
      quantity = int(line_item['quantity'])

      line_item_cancelled = False
      line_item_cancelled_at = None
      if cancelled:
        line_item_cancelled = True
        line_item_cancelled_at = order['canceled_at']
      else:
        if refunds:
          for refund in refunds:
            if refund['amount'] == total:
              line_item_cancelled = True
              line_item_cancelled_at = refund['created_at']
              break

      shipping = shippings.get(line_item['variant_id'], None)
      shipping_method = shipping['shipping_method'] if shipping else None
      shipping_charge = shipping['shipping_charge'] if shipping else None
      fulfillable_quantity = shipping['quantity'] if shipping else 0
      fulfillment_status = shipping['shipping_state'] if shipping else None

      image_url = None
      images = variant.get('images', [])
      if images:
        image_url = images[0]['product_url']

      options_text = variant['options_text']

      order_item = {
        'email': order['email'],
        # Vendor
        'vendor_id': vendor_id,
        's_order_id': order['id'],
        's_order_number': order['number'],
        's_purchase_date': order['completed_at'],
        's_total': total,
        's_price': price,
        's_tax': tax,
        's_discount': discount,
        's_quantity': quantity,
        's_fulfillable_quantity': fulfillable_quantity,
        's_fulfillment_status': fulfillment_status,
        # Closed
        # 'closed': closed,
        # 'closed_at': shopify_order.closed_at,
        's_line_item_id': line_item['id'],
        's_product_id': line_item['variant']['product_id'],
        's_variant_id': line_item['variant_id'],
        's_name': line_item['variant']['name'],
        's_sku': line_item['variant']['sku'],
        's_url': 'https://{}/products/{}'.format(self.shop_domain, variant['product_id']),
        's_image_url': image_url,
        's_options_text': options_text,
        's_shipping_method': shipping_method,
        's_shipping_charge': shipping_charge,
        # Source
        'source': source,
        'source_id': source_id,
        'canceled': line_item_cancelled,
        'cancelled_at': line_item_cancelled_at,
        'cancel_reason': None,
        's_removed': False,
        'status': order['state'],
        'note': order.get('special_instructions', None),
        'created_at': order['created_at'],
        'updated_at': order['updated_at']
      }

      order_item.update(shipping_address)
      order_item['billing_address'] = json.dumps(billing_address)

      fulfillment_data = {
        's_fulfillment_id': None,
        's_carrier': None,
        's_tracking': None,
        's_fulfillment_order_id': None,
        's_min_delivery_date': None,
        's_max_delivery_date': None,
      }
      if shipping:
        fulfillment_data = {
          's_fulfillment_id': None,
          's_carrier': shipping['carrier'],
          's_tracking': shipping['tracking'],
          's_fulfillment_order_id': shipping['shipment_id'],
          's_min_delivery_date': None,
          's_max_delivery_date': None,
        }
      order_item.update(fulfillment_data)

      risk_action = order.get('risk_action', 'accept')
      if risk_action == 'cancel':
        order_item['risky'] = True
        order_item['risky_action'] = risk_action
        order_item['risky_message'] = None
      else:
        order_item['risky'] = False
        order_item['risky_action'] = 'accept'
        order_item['risky_message'] = None

      refund_selected = None
      refund_line_item_selected = None

      order_refunds = order.get('refunds', None)
      if order_refunds:
        for order_refund in order_refunds:
          for rli in order_refund['refund_line_items']:
            if rli['line_item_id'] == line_item['id']:
              refund_selected = order_refund
              refund_line_item_selected = rli
              break

        if not refund_selected:
          for order_refund in order_refunds:
            if abs(order_refund['amount'] - line_item['total']) <= 4.99:
              refund_selected = order_refund
              break

        if not refund_selected:
          for order_refund in order_refunds:
            if order_refund['amount'] == order['total']:
              refund_selected = order_refund
              break

      if refund_selected:
        if refund_line_item_selected:
          order_item['refund_qty'] = refund_line_item_selected['quantity']
          order_item['refund_amount'] = refund_selected['amount']
        else:
          order_item['refund_qty'] = line_item['quantity']
          order_item['refund_amount'] = line_item['total']

        order_item['refund_note'] = refund_selected['refund_reason']
        order_item['refund_at'] = refund_selected['created_at']
        order_item['refund_user_id'] = None
        order_item['refunds'] = json.dumps([refund_selected])
      else:
        if order_refunds:
          amount = round(sum([float(order_refund['amount']) for order_refund in order_refunds]), 2)
          order_refund = order_refunds[0]
          order_item['refund_qty'] = 0
          order_item['refund_amount'] = amount
          order_item['refund_note'] = None
          order_item['refund_at'] = order_refund['created_at']
          order_item['refund_user_id'] = None
          order_item['refunds'] = json.dumps(order_refunds)
        else:
          order_item['refund_qty'] = 0
          order_item['refund_amount'] = 0
          order_item['refund_note'] = None
          order_item['refund_at'] = None
          order_item['refund_user_id'] = None
          order_item['refunds'] = None

      order_items.append(order_item)

    return order_items

  def capture_fulfilled_orders(self, days=21):
    today = datetime.datetime.utcnow()
    from_date = today - datetime.timedelta(days=days)

    s_order_ids = dict()
    for order_item in OrderItem.select(OrderItem).join(FulfillOrder, on=(OrderItem.id == FulfillOrder.order_item_id)).where(OrderItem.s_purchase_date > from_date).execute():
      s_order_ids[order_item.s_order_id] = None

    if not s_order_ids:
      logger.debug('[AllFulfilledOrdersCaptured] Exit')
      return

    s_order_ids = list(s_order_ids.keys())
    orders = dict()
    for order in Order.select().where((Order.s_order_id.in_(s_order_ids)) & (Order.financial_status == 'balance_due')).execute():
      orders[order.s_order_id] = order

    order_ids = list(orders.keys())
    chunks = [order_ids[x:x + 50] for x in range(0, len(order_ids), 50)]
    for chunk in chunks:
      spree_orders = self.spree_api.get_orders(chunk)
      if not spree_orders or 'error' in spree_orders or 'errors' in spree_orders:
        err_message = spree_orders.get('error', spree_orders.get('errors', '')) if spree_orders else ''
        logger.warning('[GetOrdersError] OrderIDs: %s, Error: %s', chunk, err_message)
        continue

      for spree_order in spree_orders.get('orders', []):
        if spree_order['canceled_at']:
          logger.info('[OrderCanceled] OrderNumber: %s', spree_order['number'])
          continue

        # need_capture = any([payment['state'] == 'pending' for payment in spree_order['payments']])
        # if not need_capture:
        #   continue

        order = self.spree_api.approve_order(spree_order['number'])
        paid = all([payment['state'] == 'completed' for payment in order['payments'] if payment['state'] not in ['failed', 'invalid']])
        if paid:
          logger.info('[OrderCaptured] Order: %s, OrderNumber: %s, Amount: %s', spree_order['id'], spree_order['number'], spree_order['total'])
        else:
          err_message = order.get('error', order.get('errors', '')) if order else ''
          logger.info('[OrderCaptureFailed] Order: %s, OrderNumber: %s, Error: %s', spree_order['id'], spree_order['number'], err_message)

  def upload_trackings(self, days=14):
    orders = dict()
    for order_item in OrderItem.select(OrderItem, FulfillOrder).join(FulfillOrder, on=(OrderItem.id == FulfillOrder.order_item_id), attr='fulfill_order').where((FulfillOrder.stc == 1) & (FulfillOrder.buy_tracking != '') & (FulfillOrder.buy_tracking.is_null(False)) & (OrderItem.s_fulfillment_id.is_null(True))).execute():
      if order_item.s_order_id not in orders:
        orders[order_item.s_order_id] = []
      orders[order_item.s_order_id].append(order_item)

    for order_item in OrderItem.select().where((OrderItem.s_intl_forward == 1) & (OrderItem.s_intl_tracking.is_null(False)) & (OrderItem.s_tracking.is_null(True)) & (OrderItem.s_fulfillment_id.is_null(True))).execute():
      if order_item.s_order_id not in orders:
        orders[order_item.s_order_id] = []
      orders[order_item.s_order_id].append(order_item)

    for order_item in OrderItem.select().where((OrderItem.s_tracking.is_null(False)) & (OrderItem.s_fulfillment_id.is_null(True))).execute():
      if order_item.s_order_id not in orders:
        orders[order_item.s_order_id] = []
      orders[order_item.s_order_id].append(order_item)

    for s_order_id, order_items in orders.items():
      for order_item in order_items:
        self.upload_tracking(order_item)

  def upload_tracking(self, order_item):
    if order_item.s_fulfillment_status == 'shipped' and order_item.s_tracking:
      logger.debug(
        '[OrderFulfilled] OrderNumber: %s, Tracking: %s',
        order_item.s_order_number, order_item.s_tracking)
      return

    if order_item.s_fulfillment_status == 'canceled':
      logger.debug('[ShipmentCanceled] %s', order_item.s_order_number)
      return

    quantity = order_item.s_quantity
    carrier = tracking = None
    if order_item.s_carrier:
      carrier = order_item.s_carrier
    elif order_item.s_intl_forward and order_item.s_intl_tracking:
      carrier = 'EM'
    else:
      fulfill_order = getattr(order_item, 'fulfill_order', None)
      if fulfill_order:
        carrier = fulfill_order.buy_carrier
    if order_item.s_tracking:
      tracking = order_item.s_tracking
    elif order_item.s_intl_forward and order_item.s_intl_tracking:
      tracking = order_item.s_intl_tracking
    else:
      fulfill_order = getattr(order_item, 'fulfill_order', None)
      if fulfill_order:
        tracking = fulfill_order.buy_tracking
    if not carrier or not tracking:
      return

    shipping_cost = order_item.s_shipping_cost
    shipment_id = order_item.s_fulfillment_order_id
    resp = self.spree_api.add_shipment_tracking(shipment_id, carrier, tracking, shipping_cost)
    if 'errors' in resp or 'error' in resp:
      logger.info(
        '[OrderTrackingUploadFailed] OrderNumber: %s, OrderID: %s, Carrier: %s, Tracking: %s, Error: %s',
        order_item.s_order_number, order_item.s_order_id, carrier, tracking,
        resp.get('error', resp.get('errors', '')))
    else:
      logger.info(
        '[OrderTrackingUploaded] OrderNumber: %s, ShipmentID: %s, Carrier: %s, Tracking: %s',
        order_item.s_order_number, order_item.s_fulfillment_order_id, carrier, tracking)

  def list_orders(self, params={}):
    payload = dict(params)
    if 'q[updated_at_gt]' not in params:
      if 'q[completed_at_gt]' not in params:
        days = 14
        today = datetime.datetime.utcnow()
        from_date = today - datetime.timedelta(days=days)
        from_date_s = from_date.date().isoformat()
        payload['q[completed_at_gt]'] = from_date_s

    logger.debug('[OrderFilters] %s', payload)

    if 'page' not in payload or not payload['page']:
      payload['page'] = 1

    has_more_orders = True
    while has_more_orders:
      resp = self.spree_api.list_orders(payload)
      if not resp or 'orders' not in resp or not resp['orders']:
        has_more_orders = False
        break

      for order in resp['orders']:
        if order['payment_state'] and order['payment_state'].find('refund') != -1:
          try:
            order['refunds'] = self.spree_api.get_order_refunds(order['number'])
          except Exception as e:
            logger.exception(e)

        logger.debug('[SpreeOrder] %s', order)

        yield order

      payload['page'] += 1
