import json
import datetime
import pdb

from bizdays import Calendar

from em_tasks import logger
from em_tasks.store.models import *


class OrderUtil():
  def __init__(self):
    self.calendar = Calendar()

  def audit_orders(self, days=3):
    date_format = '%Y-%m-%d'

    today = datetime.datetime.utcnow()
    from_date = (today - datetime.timedelta(days=days)).date()

    orders_activities = dict()
    query = OrderActivity.select().where(OrderActivity.event_date >= from_date)
    page = 1
    batch_size = 500
    total_cnt = query.count()
    pages = int((total_cnt + batch_size) / batch_size)
    while page <= pages:
      for order_activity in query.paginate(page, batch_size).dicts():
        if order_activity['order_item_id'] not in orders_activities:
          orders_activities[order_activity['order_item_id']] = dict()

        orders_activities[order_activity['order_item_id']][order_activity['event_name']] = order_activity

      page += 1

    order_item_ids = list(orders_activities.keys())
    n = 500
    chunks = [order_item_ids[x:x + n] for x in range(0, len(order_item_ids), n)]
    for chunk in chunks:
      orders_activities_group = dict()
      for order_item_id in chunk:
        orders_activities_group[order_item_id] = orders_activities[order_item_id]

      self.update_order_states(orders_activities_group)

  def update_order_states(self, orders_activities):
    today = datetime.datetime.utcnow()
    valid_last_update_date = self.calendar.offset(today.date(), -4)
    # valid_last_update_date = (today - datetime.timedelta(days=3)).date()
    min_domestic_tracking_date = (today - datetime.timedelta(days=5)).date()
    min_international_tracking_date = (today - datetime.timedelta(days=14)).date()
    min_fulfill_order_date = (today - datetime.timedelta(days=3)).date()
    min_domestic_fulfill_tracking_date = (today - datetime.timedelta(days=5)).date()
    min_international_fulfill_tracking_date = (today - datetime.timedelta(days=7)).date()

    order_item_ids = list(orders_activities.keys())
    order_states = dict()
    for order_state in OrderState.select().where(OrderState.order_item_id.in_(order_item_ids)).execute():
      if order_state.finished:
        orders_activities.pop(order_state.order_item_id, None)
      else:
        order_states[order_state.order_item_id] = order_state

    to_create = []
    for order_item_id, order_activities in orders_activities.items():
      order_state = order_states.get(order_item_id, None)
      order_state_data = {'order_item_id': order_item_id, 'finished': False}
      if 'OrderCancelled' in order_activities:
        order_state_data['finished'] = True
        order_state_data['state'] = 'OrderCancelled'
      elif 'OrderDelivered' in order_activities:
        order_state_data['finished'] = True
        order_state_data['state'] = 'OrderDelivered'
      elif 'OrderRefunded' in order_activities:
        order_state_data['finished'] = True
        order_state_data['state'] = 'OrderRefunded'
      elif 'OrderTrackingAdded' in order_activities:
        order_activity = order_activities['OrderTrackingAdded']
        event_data = json.loads(order_activity['event_data'])
        tracking_number = event_data.get('tracking_number', None)
        tracking_timeline = TrackingTimeline.get_or_none(TrackingTimeline.tracking_number == tracking_number)
        if tracking_timeline:
          if tracking_timeline.not_exist:
            order_state_data['state'] = 'OrderTrackingInvalid'
          else:
            if tracking_timeline.last_update_date:
              if tracking_timeline.last_update_date <= valid_last_update_date:
                order_state_data['state'] = 'OrderDeliveryDelay'
              else:
                order_state_data['state'] = 'OrderTrackingAdded'
            else:
              order_state_data['state'] = 'NoOrderTrackingStatus'
        else:
          order_state_data['state'] = 'NoOrderTrackingStatus'
      elif 'OrderPlaced' not in order_activities:
        logger.info('[NoOrderPlacedEvent] %s', order_activities)
        order_activity = list(order_activities.values())[-1]
        order_state_data['state'] = order_activity['event_name']
      else:
        if 'OrderPlaced' not in order_activities:
          logger.warning('[NoOrderPlacedEvent] %s', order_activities)
          # pdb.set_trace()
          continue

        order_activity = order_activities['OrderPlaced']
        event_data = json.loads(order_activity['event_data'])
        sku = event_data['s_sku']
        international = self.is_international_order(sku)
        if international:
          min_tracking_date = min_international_tracking_date
        else:
          min_tracking_date = min_domestic_tracking_date

        if order_activity['event_date'] < min_tracking_date:
          order_state_data['state'] = 'NoOrderTracking'
        else:
          if 'FulfillOrderDelivered' in order_activities:
            order_state_data['state'] = 'FulfillOrderDelivered'
          elif 'FulfillOrderTrackingAdded' in order_activities:
            order_state_data['state'] = 'FulfillOrderTrackingAdded'
          elif 'FulfillOrderPlaced' in order_activities:
            order_activity = order_activities['OrderPlaced']
            event_data = json.loads(order_activity['event_data'])
            sku = event_data['s_sku']
            international = self.is_international_order(sku)
            if international:
              min_fulfill_tracking_date = min_international_fulfill_tracking_date
            else:
              min_fulfill_tracking_date = min_domestic_fulfill_tracking_date
            if order_activity['event_date'] < min_fulfill_tracking_date:
              order_state_data['state'] = 'NoFulfillOrderTracking'
            else:
              order_state_data['state'] = 'FulfillOrderPlaced'
          else:
            order_activity = order_activities['OrderPlaced']
            if order_activity['event_date'] < min_fulfill_order_date:
              order_state_data['state'] = 'NoFulfillOrder'
            else:
              order_state_data['state'] = 'OrderPlaced'
            logger.debug('[OrderActivity] OrderNumber: %s, Event: %s', event_data['s_order_number'], order_activity)

      if 'state' not in order_state_data:
        logger.warning('[NoOrderState] %s', order_state_data)
        order_state_data['state'] = 'Unknown'

      if order_state:
        if order_state.state != order_state_data['state']:
          order_state.finished = order_state_data['finished']
          order_state.state = order_state_data['state']
          order_state.save()
      else:
        to_create.append(order_state_data)

    OrderState.insert_many(to_create).execute()

  def update_order_activities(self):
    order_items = dict()
    default_event_date = datetime.datetime.today()

    order_events = dict()
    batch_size = 500
    page = 1
    query = OrderItem.select().order_by(OrderItem.id.asc())
    total_cnt = query.count()
    pages = int((total_cnt + batch_size) / batch_size)
    while page <= pages:
      for order_item in query.paginate(page, batch_size).execute():
        order_items[order_item.id] = {'order_item': order_item}

      fulfill_order_trackings = dict()
      order_item_ids = list(order_items.keys())
      for fulfill_order in FulfillOrder.select().where(FulfillOrder.order_item_id.in_(order_item_ids)).execute():
        if fulfill_order.order_item_id not in order_items:
          continue

        order_item_data = order_items[fulfill_order.order_item_id]
        order_item_data['fulfill_order'] = fulfill_order
        if self.has_fulfill_order_tracking(fulfill_order):
          fulfill_order_trackings[fulfill_order.order_item_id] = {
            'tracking_number': fulfill_order.buy_tracking,
            'carrier': fulfill_order.buy_carrier
          }

      tracking_timelines = dict()
      fulfill_tracking_numbers = []
      for fulfill_order_tracking in fulfill_order_trackings.values():
        fulfill_tracking_numbers.append(fulfill_order_tracking['tracking_number'])
      # fulfill_tracking_numbers = list(set(fulfill_order_trackings.values()))
      for tracking_timeline in TrackingTimeline.select().where(TrackingTimeline.tracking_number.in_(fulfill_tracking_numbers)).execute():
        tracking_timelines[tracking_timeline.tracking_number] = tracking_timeline

      for order_item_id, fulfill_tracking in fulfill_order_trackings.items():
        order_item_data = order_items[order_item_id]
        fulfill_tracking_number = fulfill_tracking['tracking_number']

        if fulfill_tracking_number not in tracking_timelines:
          order_item_data['fulfill_tracking'] = {
            'carrier': fulfill_tracking['carrier'],
            'tracking_number': fulfill_tracking['tracking_number'],
            'delivered': None,
            'delivery_date': None
          }
          continue

        tracking_timeline = tracking_timelines[fulfill_tracking_number]
        if tracking_timeline.delivered:
          delivery_date = None
          if tracking_timeline.last_update_date:
            if not isinstance(tracking_timeline.last_update_date, str):
              try:
                delivery_date = tracking_timeline.last_update_date.isoformat()
              except Exception as e:
                logger.exception(e)
                logger.info(
                  '[TrackingTimelineError] Carrier: %s, TrackingNumber: %s, LastUpdateDate: %s',
                  tracking_timeline.carrier, tracking_timeline.tracking_number, tracking_timeline.last_update_date)

          order_item_data['fulfill_tracking'] = {
            'carrier': tracking_timeline.carrier,
            'tracking_number': fulfill_tracking_number,
            'delivered': True,
            'delivery_date': delivery_date
          }
        else:
          order_item_data['fulfill_tracking'] = {
            'carrier': tracking_timeline.carrier,
            'tracking_number': fulfill_tracking_number,
            'delivered': False,
            'delivery_date': None
          }

      trackings = dict()
      for order_item_id, order_item_data in order_items.items():
        if self.has_order_tracking(order_item_data['order_item']):
          trackings[order_item_id] = {
            'carrier': order_item_data['order_item'].s_carrier,
            'tracking_number': order_item_data['order_item'].s_tracking
          }

      tracking_numbers = []
      for tracking in trackings.values():
        tracking_numbers.append(tracking['tracking_number'])

      tracking_timelines = dict()
      for tracking_timeline in TrackingTimeline.select().where(TrackingTimeline.tracking_number.in_(tracking_numbers)).execute():
        tracking_timelines[tracking_timeline.tracking_number] = tracking_timeline

      for order_item_id, tracking in trackings.items():
        order_item_data = order_items[order_item_id]

        tracking_number = tracking['tracking_number']
        if tracking_number not in tracking_timelines:
          order_item_data['tracking'] = {
            'carrier': tracking['carrier'],
            'tracking_number': tracking['tracking_number'],
            'delivered': None,
            'delivery_date': None
          }
          continue

        tracking_timeline = tracking_timelines[tracking_number]
        if tracking_timeline.delivered:
          delivery_date = None
          if tracking_timeline.last_update_date:
            if not isinstance(tracking_timeline.last_update_date, str):
              try:
                delivery_date = tracking_timeline.last_update_date.isoformat()
              except Exception as e:
                logger.exception(e)
                logger.info(
                  '[TrackingTimelineError] Carrier: %s, TrackingNumber: %s, LastUpdateDate: %s',
                  tracking_timeline.carrier, tracking_timeline.tracking_number, tracking_timeline.last_update_date)

          order_item_data['tracking'] = {
            'carrier': tracking_timeline.carrier,
            'tracking_number': tracking_number,
            'delivered': True,
            'delivery_date': delivery_date
          }
        else:
          order_item_data['tracking'] = {
            'carrier': tracking_timeline.carrier,
            'tracking_number': tracking_number,
            'delivered': False,
            'delivery_date': None
          }

      # OrderItem and FulfillOrder prepared
      for order_item_id, order_item_entry in order_items.items():
        order_item = order_item_entry.get('order_item', None)
        if not order_item:
          logger.warning('[InvalidData] OrderItemID: %s', order_item_id)
          continue

        if order_item.id not in order_events:
          order_events[order_item.id] = []

        order_item_tracking = order_item_entry.get('tracking', None)

        order_placed_event = {
          'order_item_id': order_item.id,
          'event_name': 'OrderPlaced',
          'event_data': json.dumps({
            's_order_id': order_item.s_order_id,
            's_order_number': order_item.s_order_number,
            's_sku': order_item.s_sku,
            's_order_date': order_item.s_purchase_date.isoformat(),
            's_total': order_item.s_total
          }),
          'event_date': order_item.s_purchase_date.date(),
        }
        order_events[order_item.id].append(order_placed_event)

        if order_item.canceled:
          order_cancelled_event = {
            'order_item_id': order_item.id,
            'event_name': 'OrderCancelled',
            'event_data': json.dumps({
              's_order_id': order_item.s_order_id,
              's_order_number': order_item.s_order_number,
              'cancelled_at': order_item.cancelled_at.isoformat(),
              'cancel_reason': order_item.cancel_reason
            }),
            'event_date': order_item.cancelled_at.date()
          }
          order_events[order_item.id].append(order_cancelled_event)

        if order_item.refund_at:
          refund_at = None
          if order_item.refund_at:
            refund_at = order_item.refund_at.isoformat()
          order_refunded_event = {
            'order_item_id': order_item.id,
            'event_name': 'OrderRefunded',
            'event_data': json.dumps({
              's_order_id': order_item.s_order_id,
              's_order_number': order_item.s_order_number,
              'refunded_at': refund_at,
              'refunded_amount': order_item.refund_amount,
              'refund_qty': order_item.refund_qty,
              'refund_by': order_item.refund_user_id,
              'refund_note': order_item.refund_note
            }),
            'event_date': order_item.refund_at.date()
          }
          order_events[order_item.id].append(order_refunded_event)

        if order_item_tracking:
          event_date = default_event_date
          # TODO: Shipping label create date
          order_tracking_added_event = {
            'order_item_id': order_item.id,
            'event_name': 'OrderTrackingAdded',
            'event_data': json.dumps({
              'order_item_id': order_item.id,
              's_order_id': order_item.s_order_id,
              's_order_number': order_item.s_order_number,
              'carrier': order_item.s_carrier,
              'tracking_number': order_item.s_tracking
            }),
            'event_date': event_date
          }
          order_events[order_item.id].append(order_tracking_added_event)

          if order_item_tracking.get('delivered', False):
            event_date = default_event_date
            if order_item_tracking['delivery_date']:
              event_date = order_item_tracking['delivery_date']
            order_tracking_delivered_event = {
              'order_item_id': order_item.id,
              'event_name': 'OrderDelivered',
              'event_data': json.dumps({
                's_order_number': order_item.s_order_number,
                's_order_id': order_item.s_order_id,
                'carrier': order_item.s_carrier,
                'tracking_number': order_item.s_tracking,
                'delivery_date': order_item_tracking['delivery_date']
              }),
              'event_date': event_date
            }
            order_events[order_item.id].append(order_tracking_delivered_event)
        else:
          if order_item.s_tracking:
            logger.info('[NoTrackingUpdatesFound] Carrier: %s, TrackingNumber: %s', order_item.s_carrier, order_item.s_tracking)

        fulfill_order = order_item_entry.get('fulfill_order', None)
        fulfill_order_tracking = order_item_entry.get('fulfill_tracking', None)
        if fulfill_order:
          is_fulfill_order_valid = self.is_fulfill_order_valid(fulfill_order)
          if self.is_fulfill_order_valid(fulfill_order):
            fulfill_order_date = None
            if fulfill_order.buy_order_date:
              fulfill_order_date = fulfill_order.buy_order_date.isoformat()
            order_fulfillment = {
              's_order_number': order_item.s_order_number,
              's_order_id': order_item.s_order_id,
              'fulfill_order_number': fulfill_order.buy_order_number,
              'fulfill_order_cost': fulfill_order.buy_cost,
              'fulfill_order_date': fulfill_order_date,
              'fulfill_order_account': fulfill_order.buy_account,
            }

            event_date = default_event_date
            if order_fulfillment['fulfill_order_date']:
              event_date = order_fulfillment['fulfill_order_date'].date()

            order_fulfillment_event = {
              'order_item_id': order_item.id,
              'event_name': 'FulfillOrderPlaced',
              'event_data': json.dumps(order_fulfillment),
              'event_date': event_date,
            }
            order_events[order_item.id].append(order_fulfillment_event)

            if self.has_fulfill_order_tracking(fulfill_order):
              tracking_number = fulfill_order.buy_tracking
              carrier = fulfill_order.buy_carrier
              # TODO: Check and correct carrier
              data = {
                's_order_id': order_item.s_order_id,
                's_order_name': order_item.s_order_number,
                'fulfill_order_number': fulfill_order.buy_order_number,
                'carrier': carrier,
                'tracking_number': tracking_number,
              }

              event_date = default_event_date

              order_fulfill_tracking_event = {
                'order_item_id': order_item.id,
                'event_name': 'FulfillOrderTrackingAdded',
                'event_data': json.dumps(data),
                'event_date': event_date
              }
              order_events[order_item.id].append(order_fulfill_tracking_event)

            if fulfill_order_tracking and fulfill_order_tracking.get('delivered', False):
              event_date = default_event_date
              if fulfill_order_tracking['delivery_date']:
                event_date = fulfill_order_tracking['delivery_date']
              fulfill_order_delivered_event = {
                'order_item_id': order_item.id,
                'event_name': 'FulfillOrderDelivered',
                'event_data': json.dumps({
                  'fulfill_order_number': fulfill_order.buy_order_number,
                  'carrier': fulfill_order_tracking['carrier'],
                  'tracking_number': fulfill_order_tracking['tracking_number'],
                  'delivery_date': fulfill_order_tracking['delivery_date']
                }),
                'event_date': event_date
              }
              order_events[order_item.id].append(fulfill_order_delivered_event)
        else:
          pass

      self.process_order_events(order_events)
      order_events.clear()

      page += 1

  def process_order_events(self, order_events):
    order_item_ids = list(order_events.keys())
    for order_event in OrderActivity.select().where(OrderActivity.order_item_id.in_(order_item_ids)).execute():
      if order_event.order_item_id not in order_events:
        continue

      order_event_exist = False
      order_event_idx = None
      order_events_data = order_events[order_event.order_item_id]
      for idx, order_event_data in enumerate(order_events[order_event.order_item_id]):
        if order_event_data['event_name'] == order_event.event_name:
          order_event_exist = True
          order_event_idx = idx
          break

      if order_event_exist and order_event_idx is not None:
        order_events[order_event.order_item_id].pop(order_event_idx)

    order_events_to_create = []
    for order_item_id, order_events_data in order_events.items():
      if not order_events_data:
        continue

      order_events_to_create.extend(order_events_data)

    OrderActivity.insert_many(order_events_to_create).execute()

  def is_fulfill_order_valid(self, fulfill_order):
    # TODO: Fulfill Order policy, what could be a valid fulfill order
    return fulfill_order.buy_order_number and fulfill_order.buy_cost

  def is_fulfill_order_required(self, order_item):
    # TODO: Whether Fulfill Order is required
    return not order_item.s_sku.lower().startswith('emp')

  def has_fulfill_order_tracking(self, fulfill_order):
    carrier = fulfill_order.buy_carrier
    tracking_number = fulfill_order.buy_tracking
    if not tracking_number:
      return False

    # TODO: Check whether tracking number is valid
    return True

  def has_order_tracking(self, order_item):
    carrier = order_item.s_carrier
    tracking_number = order_item.s_tracking
    if not tracking_number:
      return False

    if not carrier or carrier.startswith('EM'):
      return False

    # TODO: Check whether tracking number is valid
    return True

  def is_delivered(self, tracking_number):
    timeline = TrackingTimeline.get_or_none(TrackingTimeline.tracking_number == tracking_number)
    if not timeline:
      return None

    return timeline.delivered

  def is_international_order(self, sku):
    sku = sku.upper()
    return sku not in ['EMP', 'EBAY_US', 'AMZ_US']

  # def is_international_order(self, order_item):
  #   sku = order_item.s_sku.upper()
  #   return sku not in ['EMP', 'EBAY_US', 'AMZ_US']

  def is_fulfill_order_required(self, order_item):
    sku = order_item.s_sku.upper()
    return sku not in ['EMP']

  def is_stc(self, order_item):
    pass
