# -*- coding: utf-8 -*-

from em_tasks import logger
from em_tasks.store.store_util import StoreUtil
from em_tasks.spree.product_util import ProductUtil
from em_tasks.utils.product_service import ProductService
from em_tasks.utils.blacklist_filter import BlacklistFilter
from em_tasks.utils.pipelines.ebay_products_filter import EbayProductsFilterPipeline
from em_tasks.utils.pipelines.price_calculator import PriceCalculatorPipeline
from em_tasks.utils.pipelines.price_filter import PriceFilterPipeline
from em_tasks.utils.pipelines.image_url_corrector import ImageUrlCorrectorPipeline
from em_tasks.utils.pipelines.upc_validator import UpcValidatorPipeline
from em_tasks.store.store_util import StoreUtil
from em_tasks.store.models import *


class EbayUploadableProductsFormatter:
  def __init__(
    self,
    store_code,
    marketplace,
    product_service,
    price_calculator,
    output_path,
    product_id_file=None,
  ):
    self.pids_buf = []
    self.batch_size = 500
    self.store = StoreUtil.get_store_by_code(store_code)
    if not self.store:
      return

    credential = json.loads(self.store.api_credential)
    self.product_util = ProductUtil(
      credential["endpoint"], credential["api_key"], credential.get("api_version", "v1")
    )

    self.blacklist_filter = BlacklistFilter()

    self.product_service = product_service
    self.marketplace = marketplace.lower()
    self.source = "Ebay_{}".format(marketplace.upper())
    self.indice_name = "ebay_{}_products".format(self.marketplace)
    self.output_path = os.path.abspath(os.path.expanduser(output_path))
    output_dir = os.path.dirname(self.output_path)
    if not os.path.isdir(output_dir):
      os.makedirs(output_dir)
    self.output_fh = open(self.output_path, "w+", encoding="utf-8", errors="ignore")

    self.blacklisted_path = os.path.abspath(
      os.path.join(
        os.path.expanduser("~"),
        ".em_tasks",
        "ebay_{}".format(self.marketplace),
        "blacklisted_products.txt",
      )
    )
    blacklisted_dir = os.path.dirname(self.blacklisted_path)
    if not os.path.isdir(blacklisted_dir):
      os.makedirs(blacklisted_dir)

    self.blacklisted_products = dict()

    self.unuploadable_path = os.path.abspath(
      os.path.join(
        os.path.expanduser("~"),
        ".em_tasks",
        "ebay_{}".format(self.marketplace),
        "unuploadable_products.txt",
      )
    )
    unuploadable_dir = os.path.dirname(self.unuploadable_path)
    if not os.path.isdir(unuploadable_dir):
      os.makedirs(unuploadable_dir)

    self.unuploadable_products = dict()

    self.products_filter_pipeline = EbayProductsFilterPipeline()

    price_calculator_pipeline = PriceCalculatorPipeline(price_calculator)
    price_filter_pipeline = PriceFilterPipeline()
    image_url_corrector_pipeline = ImageUrlCorrectorPipeline()
    upc_validator_pipeline = UpcValidatorPipeline()
    self.pipelines = [
      price_calculator_pipeline,
      price_filter_pipeline,
      image_url_corrector_pipeline,
      upc_validator_pipeline,
    ]

    self.product_id_file = product_id_file

  def run(self):
    try:
      self.load_blacklisted_products()

      logger.debug(
        "[BlacklistedProductsLoaded] ProductsCnt: %s", len(self.blacklisted_products)
      )
    except Exception as e:
      logger.exception(e)

    try:
      self.load_unuploadable_products()

      logger.debug(
        "[UnuploadableProductsLoaded] ProductsCnt: %s", len(self.unuploadable_products)
      )
    except Exception as e:
      logger.exception(e)

    try:
      self.uploaded_prods = dict()
      for prod in self.product_util.get_ebay_products(self.marketplace):
        source_product_id = prod.get("source_product_id", None)
        if not source_product_id:
          continue

        self.uploaded_prods[source_product_id] = None

      logger.debug("[UploadedProductsLoaded] ProductsCnt: %s", len(self.uploaded_prods))
    except Exception as e:
      logger.exception(e)

    try:
      options = {
        "scroll_kwargs": {"filter_path": ["hits.hits._id", "_shards", "_scroll_id"]}
      }
      if self.product_id_file:
        for prod_id, _ in self.load_product_id_from_file():
          self.process_product_id(prod_id)
      else:
        for prod_id, _ in self.product_service.load_products(self.indice_name, options):
          self.process_product_id(prod_id)

      if self.pids_buf:
        self.process_products(self.pids_buf)
        self.pids_buf = []
    finally:
      self.output_fh.close()

      self.save_blacklisted_products()
      self.save_unuploadable_products()

  def process_product_id(self, prod_id):
    if not prod_id:
      return

    if prod_id in self.blacklisted_products:
      logger.debug(
        "[ProductFiltered] ProductID: %s, Reason: Blacklisted - %s",
        prod_id,
        self.blacklisted_products[prod_id],
      )
      return

    if prod_id in self.unuploadable_products:
      logger.debug(
        "[ProductFiltered] ProductID: %s, Reason: Unuploadable - %s",
        prod_id,
        self.unuploadable_products[prod_id],
      )
      return

    if prod_id in self.uploaded_prods:
      logger.debug("[ProductFiltered] ProductID: %s, Reason: Uploaded", prod_id)
      return

    self.pids_buf.append(prod_id)
    if len(self.pids_buf) < self.batch_size:
      return

    self.process_products(self.pids_buf)
    self.pids_buf = []

  def load_product_id_from_file(self):
    with open(self.product_id_file, "r") as file:
      for line in file:
        yield line.split("\t")

  def process_products(self, pids):
    products = None
    max_retries = 3
    while max_retries > 0:
      try:
        products = self.product_service.search_products(self.indice_name, pids)
        if not isinstance(products, dict):
          logger.info(
            "[ProductService] Temporary unavailable! Wait for 10 seconds to retry."
          )
          max_retries -= 1
          time.sleep(10)
          continue

        break
      except Exception as e:
        logger.exception(e)
        max_retries -= 1
        time.sleep(3)
    if not products:
      return

    for prod_id, product in products.items():
      result = self.check_blacklist(product)
      passed = result.get("passed", True)
      reason = result.get("reason", "")
      if not passed:
        logger.debug(
          "[ProductFiltered] ProductID: %s, Reason: Blacklisted - %s", prod_id, reason
        )
        self.blacklisted_products[prod_id] = reason
        continue

      result = self.products_filter_pipeline.check_product(product)
      passed = result.get("passed", True)
      reason = result.get("reason", "")
      if not passed:
        self.unuploadable_products[prod_id] = reason
        logger.debug(
          "[ProductFiltered] ProductID: %s, Reason: Unuploadable - %s", prod_id, reason
        )
        continue

      product["source"] = self.source
      for pipeline in self.pipelines:
        product = pipeline.execute(product)
        if not product:
          break

      if not product or not product.get("price", 0):
        continue

      product["sku"] = "X02-{}".format(prod_id)

      self.output_fh.write(json.dumps(product, ensure_ascii=False))
      self.output_fh.write("\n")

  def check_blacklist(self, product):
    blacklisted = {"passed": True, "reason": ""}
    s = "{}-{}".format(
      product.get("title", "").lower(), (product.get("brand", "") or "").lower()
    )
    result = self.blacklist_filter.is_blacklisted(s)
    if result:
      blacklisted = {"passed": False, "reason": "[BlacklistKeyword] {}".format(result)}

    return blacklisted

  def load_blacklisted_products(self):
    if not os.path.isfile(self.blacklisted_path):
      return

    with open(self.blacklisted_path, encoding="utf-8", errors="ignore") as fh:
      for line in fh:
        s = line.strip()
        if not s:
          continue

        parts = s.split(",")
        pid = parts[0]
        self.blacklisted_products[pid] = parts[-1]

  def save_blacklisted_products(self):
    if not self.blacklisted_products:
      return

    records = []
    with open(self.blacklisted_path, "w+", encoding="utf-8", errors="ignore") as fh:
      for pid, reason in self.blacklisted_products.items():
        records.append("{},{}".format(pid, reason))

      fh.write("\n".join(records))
      fh.write("\n")

  def load_unuploadable_products(self):
    if not os.path.isfile(self.unuploadable_path):
      return

    with open(self.unuploadable_path, encoding="utf-8", errors="ignore") as fh:
      for line in fh:
        s = line.strip()
        if not s:
          continue

        parts = s.split(",")
        pid = parts[0]
        self.unuploadable_products[pid] = parts[-1]

  def save_unuploadable_products(self):
    if not self.unuploadable_products:
      return

    records = []
    with open(self.unuploadable_path, "w+", encoding="utf-8", errors="ignore") as fh:
      for pid, reason in self.unuploadable_products.items():
        records.append("{},{}".format(pid, reason))

      fh.write("\n".join(records))
      fh.write("\n")
