# -*- coding: utf-8 -*-

from em_tasks import logger
from em_tasks.utils.product_service import ProductService
from em_tasks.store.store_util import StoreUtil
from em_tasks.spree.product_util import ProductUtil
from em_tasks.utils.product_converters import AmzProductConverter
from em_tasks.store.models import *


class AmzUploadableProductsFilter():
  def __init__(self, store_code, marketplace, product_service, offer_service, price_caculator, output_path, asins_path=None):
    self.store = StoreUtil.get_store_by_code(store_code)
    if not self.store:
      return

    credential = json.loads(self.store.api_credential)
    self.product_util = ProductUtil(credential['endpoint'], credential['api_key'], credential.get('api_version', 'v1'))

    self.product_service = product_service
    self.offer_service = offer_service
    self.price_calc = price_caculator
    self.marketplace = marketplace.lower()
    self.pc = AmzProductConverter(self.marketplace)
    self.indice_name = 'amz_products_api_{}'.format(marketplace.lower())
    self.condition = 'new'
    self.output_path = os.path.abspath(os.path.expanduser(output_path))
    output_dir = os.path.dirname(self.output_path)
    if not os.path.isdir(output_dir):
      os.makedirs(output_dir)
    self.output_fh = open(self.output_path, 'w+', encoding='utf-8', errors='ignore')
    self.sales_rank_min = 1500000
    if self.marketplace == 'us':
      self.sales_rank_min = 500000

    self.weight_max = 2.5
    self.asins_path = asins_path

  def filter_by_file(self):
    batch_size = 500
    asin_dict = dict()
    with open(self.asins_path, "r") as file:
      for asin in file:
          asin_dict[asin.strip()] = None
          if len(asin_dict) < batch_size:
              continue
          self.filter_by_asin(list(asin_dict.keys()))
          asin_dict = {}

      if asin_dict:
          self.filter_by_asin(list(asin_dict.keys()))
          asin_dict = {}

  def filter(self):
    self.uploaded_prods = dict()
    for prod in self.product_util.get_amz_products(self.marketplace):
      source_product_id = prod.get('source_product_id', prod.get('asin', None))
      if not source_product_id:
        continue

      self.uploaded_prods[source_product_id] = None

    logger.debug('[UploadedProductsLoaded] ProductsCnt: %s', len(self.uploaded_prods))
    try:

        if self.asins_path:
          self.filter_by_file()
        else:
          self.filter_by_asin()
    finally:
      self.output_fh.close()

  def filter_by_asin(self, asins=None):
    try:
      blacklisted_path = './amz_{}_blacklisted.txt'.format(self.marketplace)
      records = dict()
      stats = dict()
      for keyword in BlacklistKeyword.select().order_by(BlacklistKeyword.id.desc()).dicts():
        k = keyword['keywords']
        stats[k] = 0

        if asins:
            query = {
              "query": {
                'query': {
                  "bool": {
                    "must": [
                      {
                        'multi_match': {
                          'query': k,
                          'type': 'phrase',
                          'fields': ['brand^10', 'title']
                        }
                      },
                      {
                        "terms": {
                          "asin": asins
                        }
                      }
                    ]
                  }
                }
              }
            }
        else:
            query = {
              'query': {
                'query': {
                  'multi_match': {
                    'query': k,
                    'type': 'phrase',
                    'fields': ['brand^10', 'title']
                  }
                }
              }
            }
        for prod_id, _ in self.product_service.load_products(self.indice_name, query):
          records[prod_id] = None
          stats[k] += 1

      total_products_cnt = 0
      with open(blacklisted_path, 'w+', encoding='utf-8', errors='ignore') as fh:
        fh.write("\n".join(records.keys()))

      for k, products_cnt in stats.items():
        if not products_cnt:
          continue

        logger.debug('[BlacklistedProducts] Keyword: %s, ProductCnt: %s', k, products_cnt)

        total_products_cnt += products_cnt

      logger.debug('[BlacklistedProductsLoaded] ProductsCnt: %s', total_products_cnt)



      if asins:
          query = {"query": {"query": {"bool": {"filter": [{"terms": {"asin": asins}}]}}}}
      else:
        query = None

      prods_buf = dict()
      batch_size = 500
      for prod_id, prod in self.product_service.load_products(self.indice_name, query):
        exist = prod.get('existence', True)
        if not exist:
          continue

        has_only_default_variant = prod.get('has_only_default_variant', True)
        if not has_only_default_variant:
          continue

        if prod_id in records or prod_id in self.uploaded_prods:
          continue

        weight = self.pc.get_weight(prod)
        if weight and weight > self.weight_max:
          continue

        is_fragile = self.pc.is_fragile(prod)
        if is_fragile:
          continue

        sales_rank = self.pc.get_sales_rank(prod)
        if not sales_rank or sales_rank > self.sales_rank_min:
          continue

        prods_buf[prod_id] = prod
        if len(prods_buf) < batch_size:
          continue

        self.process_products(prods_buf)
        prods_buf.clear()

      if prods_buf:
        self.process_products(prods_buf)
    except Exception as e:
        pass

  def process_products(self, prods):
    asins = list(prods.keys())
    offers = dict()
    src_offers = self.offer_service.get_offers(self.marketplace, asins, self.condition)
    if not src_offers:
      return

    for asin, product_info in prods.items():
      prod = self.pc.convert(product_info)
      src_offer = src_offers.get(asin, None)
      offer = self.price_calc.calc_offer(src_offer)
      if not offer:
        continue

      if offer['quantity'] <= 0 or offer['price'] <= 0:
        continue

      prod['price'] = offer['price']
      prod['quantity'] = offer['quantity']
      prod['cost_price'] = src_offer.get('price', 0)

      self.output_fh.write("{}\n".format(json.dumps(prod)))
