# -*- coding: utf-8 -*-

import datetime
import time
import pdb

import dateutil
import dateparser

from tracking_tracker.dhl_api import DHLApi

from em_tasks import logger
from em_tasks.store.store_util import StoreUtil
from em_tasks.store.models import *


class DHLTrackingTracker():
  def __init__(self, api_key, environment='production', days=30):
    self.days = days
    self.api_key = api_key
    self.environment = environment
    self.api = DHLApi(api_key, environment)

  def run(self):
    carrier = 'DHL'
    carrier_pattern = '%DHL%'
    now = datetime.datetime.now()
    from_date = now - datetime.timedelta(days=self.days)
    trackings = dict()
    for order_item in OrderItem.select(OrderItem.id, OrderItem.s_carrier, OrderItem.s_tracking).where((OrderItem.s_purchase_date >= from_date) & (OrderItem.s_carrier ** carrier_pattern)).execute():
      if not order_item.s_carrier or not order_item.s_tracking:
        continue

      trackings[order_item.s_tracking] = None

    for fulfill_order in FulfillOrder.select(FulfillOrder.id, FulfillOrder.buy_tracking).where(FulfillOrder.buy_carrier ** carrier_pattern).execute():
      if not fulfill_order.buy_tracking:
        continue

      trackings[fulfill_order.buy_tracking] = None

    for customer_return in CustomerReturn.select(CustomerReturn.id, CustomerReturn.tracking_number).where(CustomerReturn.carrier ** carrier_pattern).execute():
      if not customer_return.tracking_number:
        continue

      trackings[customer_return.tracking_number] = None

    tracking_numbers = list(trackings.keys())
    for tracking_timeline in TrackingTimeline.select().where(TrackingTimeline.tracking_number.in_(tracking_numbers)).execute():
      if tracking_timeline.delivered or tracking_timeline.not_exist:
        trackings.pop(tracking_timeline.tracking_number, None)
        continue

    if not trackings:
      return

    for tracking_number in trackings.keys():
      max_retries = 3
      while max_retries > 0:
        try:
          tracking = self.api.get_tracking_details(tracking_number)
          if not tracking or 'shipments' not in tracking:
            if 'status' in tracking and tracking['status'] == 429:
              logger.debug('[TooManyTrackingRequests] TrackingNumber: %s, Response: %s', tracking_number, tracking)

              max_retries -= 1

              time.sleep(7)

              continue

            tracking_timeline = TrackingTimeline.get_or_none(TrackingTimeline.tracking_number == tracking_number)
            if not tracking_timeline:
              tracking_timeline = TrackingTimeline()
              tracking_timeline.carrier = carrier
              tracking_timeline.tracking_number = tracking_number

            tracking_timeline.not_exist = True
            tracking_timeline.save()

            continue

          timelines = self.parse_timelines(tracking)
          delivered = any([timeline['status'].lower().find('delivered') != -1 for timeline in timelines if timeline['status']])
          status = 'IN_TRANSIT'
          if delivered:
            status = 'DELIVERED'

          last_update_date = None
          if timelines:
            last_update = timelines[0]
            date_s = last_update['date']
            parsed_date = dateparser.parse(date_s)
            if parsed_date:
              last_update_date = parsed_date.strftime('%Y-%m-%d')
            else:
              last_update_date = date_s

          tracking_timeline = TrackingTimeline.get_or_none(TrackingTimeline.tracking_number == tracking_number)
          if tracking_timeline:
            tracking_timeline.delivered = delivered
            tracking_timeline.last_update_date =  last_update_date
            tracking_timeline.timelines = json.dumps(timelines)
            tracking_timeline.save()
          else:
            data = {
              'carrier': carrier,
              'tracking_number': tracking_number,
              'delivered': delivered,
              'last_update_date': last_update_date,
              'timelines': json.dumps(timelines)
            }
            TrackingTimeline.insert(data).execute()
            logger.info('[TrackingTimeline] %s', data)
        except Exception as e:
          logger.exception(e)

  def parse_timelines(self, tracking):
    if not tracking or 'shipments' not in tracking or not tracking['shipments']:
      return []

    timelines = []
    d = dict()
    shipment = tracking['shipments'][0]
    for event in shipment.get('events', []):
      date_s = event['timestamp']
      parsed_date = None
      try:
        parsed_date = dateutil.parser.parse(date_s)
      except:
        try:
          parsed_date = dateparser.parse(date_s)
        except Exception as e:
          # logger.exception(e)
          parsed_date = None
      if parsed_date:
        date_s = parsed_date.strftime('%m/%d/%Y')

      location = event.get('location', d).get('address', d).get('addressLocality', None)

      description = event.get('description', '')
      status = 'IN_TRANSIT'
      if description.lower().find('delivered') != -1:
        status = 'Delivered'

      timelines.append({
        'status': status,
        'status_detail': description,
        'location': location,
        'date': date_s
      })

    return timelines
