import logging
from datetime import datetime
import time

from pymongo import MongoClient, UpdateOne
from pymongo.errors import (
    BulkWriteError,
    ConnectionFailure,
    OperationFailure,
    ServerSelectionTimeoutError,
)


class ErrorService:
    MAX_RETRY = 3

    def __init__(self, host, port, user, password):
        mongo_uri = f"mongodb://{user}:{password}@{host}:{port}/?authSource=admin"
        self.client = MongoClient(mongo_uri)
        self.mongo_db = self.client["errors"]

    def save_no_product_info_asins(self, collection, asins, marketplace):
        return self.save_no_offer_asins(collection, asins, marketplace)

    def save_no_offer_asins(self, collection, asins, marketplace):
        items = [
            UpdateOne(
                {"_id": f"{marketplace.lower()}-{asin}"},
                {"$set":
                    {
                        "asin": str(asin), "marketplace": marketplace.lower(), "time": datetime.now().replace(microsecond=0)
                    }
                },
                upsert=True
            ) for asin in asins]

        retry_left_count = self.MAX_RETRY
        while retry_left_count > 0:
            try:
                self.mongo_db[collection.lower()].bulk_write(items)
                logging.info(f"[Items Saved]: {items}")
                break
            except BulkWriteError as bwe:
                for error in bwe.details["writeErrors"]:
                    logging.exception(f"{error['errmsg']}")
                break

            except OperationFailure as of:
                logging.exception(f"{str(of)}")
                break

            except (ConnectionFailure, ServerSelectionTimeoutError) as e:
                time.sleep(2 ** (self.MAX_RETRY - retry_left_count))
                retry_left_count = retry_left_count - 1

            except Exception as e:
                raise e


