# -*- coding: utf-8 -*-

import datetime
import pdb

import dateutil
import dateparser

from tracking_tracker.fedex_api import FedexApi

from em_tasks import logger
from em_tasks.store.store_util import StoreUtil
from em_tasks.store.models import *


class FedexTrackingTracker():
  def __init__(self, client_id, client_secret, environment='production', grant_type='client_credentials', days=30):
    self.days = days
    self.client_id = client_id
    self.client_secret = client_secret
    self.environment = environment
    self.grant_type = grant_type
    self.api = FedexApi(client_id, client_secret, environment, grant_type)

  def run(self):
    carrier = 'Fedex'
    carrier_pattern = '%fedex%'
    now = datetime.datetime.now()
    from_date = now - datetime.timedelta(days=self.days)
    trackings = dict()
    for order_item in OrderItem.select(OrderItem.id, OrderItem.s_carrier, OrderItem.s_tracking).where((OrderItem.s_purchase_date >= from_date) & (OrderItem.s_carrier ** carrier_pattern)).execute():
      if not order_item.s_carrier or not order_item.s_tracking:
        continue

      trackings[order_item.s_tracking] = None

    for fulfill_order in FulfillOrder.select(FulfillOrder.id, FulfillOrder.buy_tracking).where(FulfillOrder.buy_carrier ** carrier_pattern).execute():
      if not fulfill_order.buy_tracking:
        continue

      trackings[fulfill_order.buy_tracking] = None

    for customer_return in CustomerReturn.select(CustomerReturn.id, CustomerReturn.tracking_number).where(CustomerReturn.carrier ** carrier_pattern).execute():
      if not customer_return.tracking_number:
        continue

      trackings[customer_return.tracking_number] = None

    tracking_numbers = list(trackings.keys())
    for tracking_timeline in TrackingTimeline.select().where(TrackingTimeline.tracking_number.in_(tracking_numbers)).execute():
      if tracking_timeline.delivered or tracking_timeline.not_exist:
        trackings.pop(tracking_timeline.tracking_number, None)
        continue

    if not trackings:
      return

    for tracking_number in trackings.keys():
      try:
        tracking = self.api.get_tracking_details(tracking_number)
        if not tracking or 'output' not in tracking or 'completeTrackResults' not in tracking['output'] or not tracking['output']['completeTrackResults']:
          logger.info('[InvalidResponse] %s', tracking)
          continue

        track_results = None
        for complete_track_result in tracking['output']['completeTrackResults']:
          tn =  complete_track_result.get('trackingNumber', None)
          if tn == tracking_number:
            track_results = complete_track_result.get('trackResults', None)
            break

        if not track_results:
          logger.info('[NoTrackResults] %s', tracking_number)
          continue

        valid_track_result = None
        for track_result in track_results:
          if 'error' in track_result and track_result['error'].get('code', None) == 'TRACKING.TRACKINGNUMBER.NOTFOUND':
            tracking_timeline = TrackingTimeline.get_or_none(TrackingTimeline.tracking_number == tracking_number)
            if not tracking_timeline:
              tracking_timeline = TrackingTimeline()
              tracking_timeline.carrier = carrier
              tracking_timeline.tracking_number = tracking_number

            tracking_timeline.not_exist = True
            tracking_timeline.save()

            continue

          valid_track_result = track_result

        if not valid_track_result:
          tracking_timeline = TrackingTimeline.get_or_none(TrackingTimeline.tracking_number == tracking_number)
          if not tracking_timeline:
            tracking_timeline = TrackingTimeline()
            tracking_timeline.carrier = carrier
            tracking_timeline.tracking_number = tracking_number

          tracking_timeline.not_exist = True
          tracking_timeline.save()

          logger.info('[NoValidTrackResult] %s', tracking_number)

          continue

        timelines = self.parse_timelines(valid_track_result)
        delivered = any([timeline['status'].lower().find('delivered') != -1 for timeline in timelines if timeline['status']])
        status = 'IN_TRANSIT'
        if delivered:
          status = 'DELIVERED'

        last_update_date = None
        if timelines:
          last_update = timelines[0]
          date_s = last_update['date']
          parsed_date = dateparser.parse(date_s)
          if parsed_date:
            last_update_date = parsed_date.strftime('%Y-%m-%d')
          else:
            last_update_date = date_s

        tracking_timeline = TrackingTimeline.get_or_none(TrackingTimeline.tracking_number == tracking_number)
        if tracking_timeline:
          tracking_timeline.delivered = delivered
          tracking_timeline.last_update_date =  last_update_date
          tracking_timeline.timelines = json.dumps(timelines)
          tracking_timeline.save()
        else:
          data = {
            'carrier': carrier,
            'tracking_number': tracking_number,
            'delivered': delivered,
            'last_update_date': last_update_date,
            'timelines': json.dumps(timelines)
          }
          TrackingTimeline.insert(data).execute()
          logger.info('[TrackingTimeline] %s', data)
      except Exception as e:
        logger.exception(e)

  def parse_timelines(self, tracking):
    if not tracking or 'scanEvents' not in tracking or not tracking['scanEvents']:
      return []

    timelines = []
    d = dict()
    shipment = tracking['scanEvents']
    for event in tracking['scanEvents']:
      date_s = event['date']
      parsed_date = None
      try:
        parsed_date = dateutil.parser.parse(date_s)
      except:
        try:
          parsed_date = dateparser.parse(date_s)
        except Exception as e:
          # logger.exception(e)
          parsed_date = None
      if parsed_date:
        date_s = parsed_date.strftime('%m/%d/%Y')

      scan_location = event.get('scanLocation', d)
      location = '{}, {} {}'.format(scan_location.get('city', ''), scan_location.get('stateOrProvinceCode', ''), scan_location.get('postalCode', ''))

      status = event.get('derivedStatus', '')
      description = event.get('eventDescription', '')

      timelines.append({
        'status': status,
        'status_detail': description,
        'location': location,
        'date': date_s
      })

    return timelines
