from em_tasks.utils.pipelines.base import BasePipeline
from em_tasks.utils.product_converters import AmzProductConverter

from em_tasks import logger


class AmzProductsFilterPipeline(BasePipeline):
  def __init__(self, marketplace = 'us', weight_max = 2.5, dimension_max = 30, sales_rank_min = 1500000, export = True):
    self.weight_max = weight_max
    self.dimension_max = dimension_max
    self.sales_rank_min = sales_rank_min
    self.export = export
    self.marketplace = marketplace.lower()
    self.pc = AmzProductConverter(self.marketplace)
    self.source = 'AMZ_{}'.format(marketplace.upper())

  def execute(self, product):
    result = self.check_product(product)
    passed = result.get('passed', True)
    if passed:
      product['source'] = self.source
      return product

  def check_product(self, product):
    exist = product.get('existence', True)
    if not exist:
      return {'passed': False, 'reason': '[NotExist]'}

    has_only_default_variant = product.get('has_only_default_variant', True)
    if not has_only_default_variant:
      return {'passed': False, 'reason': '[MultipleVariant]'}

    title = product.get('title', '').strip().lower()
    if not title:
      return {'passed': False, 'reason': '[NoTitle]'}

    if title.find('heavy') != -1:
      return {'passed': False, 'reason': '[OverWeight]'}

    if title.find('large') != -1:
      return {'passed': False, 'reason': '[OverSize]'}

    weight = self.pc.get_weight(product)
    if weight and weight > self.weight_max:
      return {'passed': False, 'reason': '[OverWeight] {}'.format(weight)}

    is_fragile = self.pc.is_fragile(product)
    if is_fragile:
      return {'passed': False, 'reason': '[Fragile]'}

    is_alcohol = self.pc.is_alcohol(product)
    if is_alcohol:
      return {'passed': False, 'reason': '[Alcohol]'}

    is_contains_food_or_beverage = self.pc.is_contains_food_or_beverage(product)
    if is_contains_food_or_beverage:
      return {'passed': False, 'reason': '[FoodOrBeverage]'}

    is_grow_your_own_kits = self.pc.is_grow_your_own_kits(product)
    if is_grow_your_own_kits:
      return {'passed': False, 'reason': '[GrowYourOwnKits]'}

    is_sensitive = self.pc.is_sensitive(product)
    if is_sensitive:
      return {'passed': False, 'reason': '[Sensitive]'}

    # is_liquid = self.pc.is_liquid(product)
    # if is_liquid:
    #   logger.debug('[Liquid] ASIN: %s, Title: %s', product.get('asin', ''), product.get('title', ''))

    is_plant_or_seed = self.pc.is_plant_or_seed(product)
    if is_plant_or_seed:
      return {'passed': False, 'reason': '[PlantOrSeed]'}

    if self.export:
      is_spray = self.pc.is_spray(product)
      if is_spray:
        return {'passed': False, 'reason': '[Spray]'}

      is_test_kit = self.pc.is_test_strip(product)
      if is_test_kit:
        return {'passed': False, 'reason': '[TestKit]'}

      is_batteries_included = self.pc.is_batteries_included(product)
      if is_batteries_included:
        return {'passed': False, 'reason': '[IncludeBatteries]'}

    dimensions = self.pc.get_dimensions(product)
    if dimensions:
      width = dimensions.get('width', None)
      height = dimensions.get('height', None)
      length = dimensions.get('length', None)
      if (width and width > self.dimension_max) or (height and height > self.dimension_max) or (length and length > self.dimension_max):
        return {'passed': False, 'reason': '[OverSize]'}

    sales_rank = self.pc.get_sales_rank(product)
    if sales_rank and sales_rank > self.sales_rank_min:
      return {'passed': False, 'reason': '[SalesRankLow]'}

    return {'passed': True, 'reason': ''}
