import os
import time

from em_tasks.utils.product_converters import AmzProductConverter
from em_tasks.utils.product_loaders.base import BaseProductLoader
from em_tasks import logger


class AmzEsProductLoader(BaseProductLoader):
  def __init__(self, product_service, offer_service, marketplace, product_ids_path):
    self.product_service = product_service
    self.offer_service = offer_service
    self.marketplace = marketplace.lower()
    self.indice_name = 'amz_products_api_{}'.format(self.marketplace)
    self.condition = 'new'
    self.pc = AmzProductConverter(self.marketplace)
    self.product_ids_path = None
    if product_ids_path:
      self.product_ids_path = os.path.abspath(os.path.expanduser(product_ids_path))

  def load_products_by_id(self):
    if not self.product_ids_path:
      return

    batch_size = 500
    product_ids_buf = []
    with open(self.product_ids_path, encoding='utf-8', errors='ignore') as fh:
      for line in fh:
        s = line.strip()
        if not s:
          continue

        product_id = s
        product_ids_buf.append(product_id)
        if len(product_ids_buf) < batch_size:
          continue

        products = self.get_products_by_id(product_ids_buf)
        if not products:
          continue

        for product in products:
          yield product

        product_ids_buf = []

    if product_ids_buf:
      products = self.get_products_by_id(product_ids_buf)
      if not products:
        return

      for product in products:
        yield product

      product_ids_buf = []

  def load(self):
    if not self.product_ids_path:
      return

    batch_size = 500
    product_ids_buf = []
    with open(self.product_ids_path, encoding='utf-8', errors='ignore') as fh:
      for line in fh:
        s = line.strip()
        if not s:
          continue

        product_id = s
        product_ids_buf.append(product_id)
        if len(product_ids_buf) < batch_size:
          continue

        products = self.get_products(product_ids_buf)
        if not products:
          continue

        for product in products:
          yield product

        product_ids_buf = []

    if product_ids_buf:
      products = self.get_products(product_ids_buf)
      if not products:
        return

      for product in products:
        yield product

      product_ids_buf = []

  def get_products_by_id(self, product_ids):

    max_retries = 3
    while max_retries > 0:
      try:
        products = self.product_service.search_products(self.indice_name, product_ids)
        if not isinstance(products, dict):
          logger.info('[ProductService] Temporary unavailable! Wait for 10 seconds to retry.')
          max_retries -= 1
          time.sleep(10)
          continue

        for product_id in product_ids:
          if product_id not in products or not products[product_id]:
            logger.debug('[NoProductInformation] ASIN: %s', product_id)

        for product_id, product_info in products.items():
            yield product_id, product_info
        break
      except Exception as e:
        logger.exception(e)
        max_retries -= 1
        time.sleep(3)

  def get_products(self, product_ids):
    src_offers = None
    max_retries = 3
    while max_retries > 0:
      try:
        src_offers = self.offer_service.get_offers(self.marketplace, product_ids, self.condition)
        if not isinstance(src_offers, dict):
          logger.info('[OfferService] Temporary unavailable! Wait for 10 seconds to retry.')
          max_retries -= 1
          time.sleep(10)
          continue

        break
      except Exception as e:
        logger.exception(e)
        max_retries -= 1
        time.sleep(3)
    if not src_offers:
      return

    items = []
    max_retries = 3
    while max_retries > 0:
      try:
        products = self.product_service.search_products(self.indice_name, product_ids)
        if not isinstance(products, dict):
          logger.info('[ProductService] Temporary unavailable! Wait for 10 seconds to retry.')
          max_retries -= 1
          time.sleep(10)
          continue

        for product_id in product_ids:
          if product_id not in products or not products[product_id]:
            logger.debug('[NoProductInformation] ASIN: %s', product_id)

        for product_id, product_info in products.items():
          product = self.pc.convert(product_info)
          src_offer = src_offers.get(product_id, None)
          if not src_offer:
            logger.debug('[NoOffer] ASIN: %s', product_id)
            continue

          product['price'] = src_offer.get('price', 0)
          product['currency'] = src_offer.get('currency', 'USD')

          items.append(product)

        break
      except Exception as e:
        logger.exception(e)
        max_retries -= 1
        time.sleep(3)

    return items
