import json
import math

from shopping_utils.google_shopping import GoogleShopping

from em_tasks import logger
from em_tasks.shopify.product_util import ProductUtil
from em_tasks.store.store_util import StoreUtil
from em_tasks.store.models import *


class ShopifyAmzCustomLabelUpdater():
  def __init__(self, store_code, marketplace, offer_service, price_caculator):
    self.condition = 'new'

    self.store = StoreUtil.get_store_by_code(store_code)
    if not self.store:
      return

    credential = json.loads(self.store.api_credential)
    self.product_util = ProductUtil(credential['name'], credential['token'], credential.get('release', None))

    self.marketplace = marketplace
    self.offer_service = offer_service
    self.price_calc = price_caculator

    self.gss = dict()
    gmc_stores = GmcStore.select().execute()
    for gmc_store in gmc_stores:
      self.gss[gmc_store.id] = GoogleShopping(
        gmc_store.merchant_id, json.loads(gmc_store.service_account_data))

  def run(self):
    batch_size = 250
    prods = dict()
    for amz_prod in StoreUtil.get_amz_products(self.marketplace):
      prods[amz_prod['product_id']] = amz_prod
      if len(prods) < batch_size:
        continue

      self.process_products(prods)
      prods.clear()

    if prods:
      self.process_products(prods)
      prods.clear()

  def process_products(self, prods):
    asins = [prod['asin'] for prod in prods.values()]
    src_offers = self.offer_service.get_offers(self.marketplace, asins, self.condition)
    if not src_offers:
      return

    profit_rates = dict()
    for asin, src_offer in src_offers.items():
      profit_rate = self.price_calc.calc_profit_rate(src_offer)
      if not profit_rate:
        continue

      profit_rates[asin] = profit_rate

    prod_profit_rates = dict()
    for _, prod in prods.items():
      if 'handle' not in prod or not prod['handle']:
        continue

      asin = prod['asin']
      if asin not in profit_rates:
        continue

      profit_rate = profit_rates[asin]
      prod_profit_rates[prod['handle']] = profit_rate

    self.update_gmc_custom_label(prod_profit_rates)

  def update_gmc_custom_label(self, profit_rates):
    gmc_entries = dict()
    handles = list(profit_rates.keys())
    gmc_offers = dict()
    for gmc_product in GmcProduct.select().where(GmcProduct.handle.in_(handles)).execute():
      if gmc_product.handle not in profit_rates:
        continue

      if gmc_product.gmc_store_id not in gmc_entries:
        gmc_entries[gmc_product.gmc_store_id] = []

      profit_rate = profit_rates[gmc_product.handle]
      entry = {
        'productId': gmc_product.gmc_id,
        'customLabel0': self.format_profit_rate(profit_rate),
      }
      gmc_entries[gmc_product.gmc_store_id].append(entry)

    for gmc_store_id, entries in gmc_entries.items():
      if gmc_store_id not in self.gss:
        continue

      if not entries:
        continue

      gs = self.gss[gmc_store_id]
      result = gs.update_products(entries)

      logger.info(
        '[GoogleMerchantCustomLabelUpdate] Succeed: %s, Failed: %s',
        result['succeed_cnt'], result['failed_cnt'])

  def format_profit_rate(self, profit_rate):
    profit_level = math.floor(profit_rate / 10)
    return '{}-{}%'.format(profit_level * 10 + 1, (profit_level + 1) * 10)
