# -*- coding: utf-8 -*-

import pdb
import json

from shopping_utils.google_shopping import GoogleShopping

from em_tasks import logger
from em_tasks.store.store_util import StoreUtil
from em_tasks.shopify.product_util import ProductUtil
from em_tasks.store.models import *


class ShopifyCollectiveSyncer():
  def __init__(self, store_code, merchant_id, next_page_url=None):
    self.store = StoreUtil.get_store_by_code(store_code)
    if not self.store:
      return

    credential = json.loads(self.store.api_credential)
    self.product_util = ProductUtil(credential['name'], credential['token'], credential.get('release', None))
    self.next_page_url = next_page_url
    self.source = 'Collective'
    self.store_seller_id = 1
    self.merchant_id = merchant_id
    gmc_store = GmcStore.get_or_none(GmcStore.merchant_id == merchant_id)
    self.gs = GoogleShopping(
      gmc_store.merchant_id, json.loads(gmc_store.service_account_data))
    self.product_ids = []

  def run(self):
    products_buf = dict()
    batch_size = 1000
    next_page_url = self.next_page_url
    max_retries = 12
    while max_retries > 0:
      try:
        for products, next_page_url in self.product_util.list_products(next_page_url, **{'status': 'draft'}):
          for product in products:
            if product.tags.lower().find('collective') == -1:
              continue

            products_buf[product.id] = product
            self.product_ids.append(product.id)

          if len(products_buf) < batch_size:
            continue

          self.process_products(products_buf)
          products_buf.clear()

        break
      except Exception as e:
        logger.exception(e)
        max_retries -= 1

    if products_buf:
      self.process_products(products_buf)
      products_buf.clear()

    self.publish_products()

  def process_products(self, products):
    to_create = []
    to_update = []
    product_ids = list(products.keys())
    for sp in StoreProduct.select().where(StoreProduct.s_product_id.in_(product_ids)).execute():
      if sp.s_product_id not in products:
        continue

      product = products.pop(sp.s_product_id)
      in_stock = False
      for v in product.variants:
        if v.inventory_quantity > 0:
          in_stock = True
          break

      variant = product.variants[0]
      record = {
        'id': sp.id,
        's_product_id': product.id,
        'handle': product.handle,
        'title': product.title,
        'description': product.body_html,
        'has_image': bool(product.images),
        'sku': variant.sku,
        'in_stock':in_stock,
        'source': self.source,
        'source_product_id': product.id
      }
      to_update.append(record)

    if to_update:
      StoreProduct.insert_many(to_update).on_conflict(
        preserve=(StoreProduct.handle, StoreProduct.title, StoreProduct.description, StoreProduct.has_image, StoreProduct.sku, StoreProduct.in_stock)).execute()
      to_update = []

    if not products:
      return

    for product_id, product in products.items():
      in_stock = False
      for v in product.variants:
        if v.inventory_quantity > 0:
          in_stock = True
          break

      variant = product.variants[0]
      record = {
        'store_seller_id': self.store_seller_id,
        's_product_id': product.id,
        'handle': product.handle,
        'title': product.title,
        'description': product.body_html,
        'has_image': bool(product.images),
        'sku': variant.sku,
        'in_stock': in_stock,
        'source': self.source,
        'source_product_id': product.id
      }
      to_create.append(record)

    StoreProduct.insert_many(to_create).execute()

  def publish_products(self):
    chunks = [self.product_ids[x:x + 250] for x in range(0, len(self.product_ids), 250)]
    for chunk in chunks:
      products = self.product_util.publish_products(chunk)
      gmc_products = []
      for _, product in products.items():
        gmc_product = self.product_util.to_gmc_product(product)
        if not gmc_product:
          continue
        gmc_products.append(gmc_product)

      result = self.gs.upload_products(gmc_products)
      logger.info('[GmcUpload] %s', result)
