# -*- coding: utf-8 -*-

import datetime
import json
import time
import re
import math

import click
import dateparser

from shopping_utils.google_shopping import GoogleShopping

from em_tasks import logger
from em_tasks.shopify.product_util import ProductUtil
from em_tasks.store.store_util import StoreUtil
from em_tasks.store.models import *


class ShopifyEbayCustomLabelUpdater():
  def __init__(self, store_code, product_service, price_calculator, marketplace, ttl=7):
    self.store = StoreUtil.get_store_by_code(store_code)
    if not self.store:
      return

    credential = json.loads(self.store.api_credential)
    self.product_util = ProductUtil(credential['name'], credential['token'], credential.get('release', None))

    self.product_service = product_service
    self.price_calc = price_calculator
    self.marketplace = marketplace
    self.indice_name = 'ebay_{}_products'.format(marketplace.lower())
    today = datetime.date.today()
    self.expire_date = today - datetime.timedelta(days=ttl) 

    self.gss = dict()
    gmc_stores = GmcStore.select().execute()
    for gmc_store in gmc_stores:
      self.gss[gmc_store.id] = GoogleShopping(
        gmc_store.merchant_id, json.loads(gmc_store.service_account_data))

  def run(self):
    batch_size = 250
    products = dict()
    for ebay_prod in StoreUtil.get_ebay_products(self.marketplace):
      products[ebay_prod['source_product_id']] = ebay_prod

      if len(products) < batch_size:
        continue

      self.process_products(products)
      products.clear()

    if products:
      self.process_products(products)
      products = []

  def process_products(self, prods):
    product_ids = list(prods.keys())
    products = False
    max_retries = 3
    while max_retries > 0:
      try:
        products = self.product_service.search_products(self.indice_name, product_ids)
        if not isinstance(products, dict):
          logger.info('[ProductService] Temporary unavailable! Wait for 10 seconds to retry.')
          max_retries -= 1
          time.sleep(10)
          continue

        break
      except Exception as e:
        logger.exception(e)
        max_retries -= 1
        time.sleep(3)

    if not products:
      return

    profit_rates = dict()
    for pid, product in products.items():
      if pid not in prods:
        continue

      prod = prods[pid]

      src_offer = None

      existence = product.get('existence', True)
      has_only_default_variant = product.get('has_only_default_variant', True)
      src_product_price = product.get('price', 0)
      src_shipping_fee = product.get('shipping_fee', 0) or 0
      currency = product.get('currency', None) or 'USD'
      if not existence or not has_only_default_variant or not src_product_price:
        prod['offer'] = {'price': 0, 'quantity': 0}
      else:
        update_date_s = product.get('date', product.get('time', None))
        if not update_date_s:
          continue

        update_date = None
        try:
          update_date = dateparser.parse(update_date_s).date()
        except:
          pass

        if not update_date:
          continue

        shipping_days_max = product.get('shipping_days_max', None)
        shipping_invalid = not shipping_days_max or shipping_days_max > 14
        if not existence or not has_only_default_variant or not src_product_price or shipping_invalid:
          # Out of stock
          src_offer = {'price': src_product_price + src_shipping_fee, 'currency': currency}
        else:
          src_offer = {'price': src_product_price + src_shipping_fee, 'currency': currency}

        offer = self.price_calc.calc_offer(src_offer)
        profit_rate = self.price_calc.calc_profit_rate(src_offer)
        if not profit_rate:
          continue

        profit_rates[pid] = profit_rate

    prod_profit_rates = dict()
    for _, prod in prods.items():
      if 'handle' not in prod or not prod['handle']:
        continue

      source_product_id = prod['source_product_id']
      if source_product_id not in profit_rates or not profit_rates[source_product_id]:
        continue

      profit_rate = profit_rates[source_product_id]
      prod_profit_rates[prod['handle']] = profit_rate

    self.update_gmc_custom_label(prod_profit_rates)

  def update_gmc_custom_label(self, profit_rates):
    gmc_entries = dict()
    handles = list(profit_rates.keys())
    gmc_offers = dict()
    for gmc_product in GmcProduct.select().where(GmcProduct.handle.in_(handles)).execute():
      if gmc_product.handle not in profit_rates:
        continue

      if gmc_product.gmc_store_id not in gmc_entries:
        gmc_entries[gmc_product.gmc_store_id] = []

      profit_rate = profit_rates[gmc_product.handle]
      entry = {
        'productId': gmc_product.gmc_id,
        'customLabel0': self.format_profit_rate(profit_rate),
      }
      gmc_entries[gmc_product.gmc_store_id].append(entry)

    for gmc_store_id, entries in gmc_entries.items():
      if gmc_store_id not in self.gss:
        continue

      if not entries:
        continue

      gs = self.gss[gmc_store_id]
      result = gs.update_products(entries)

      logger.info(
        '[GoogleMerchantCustomLabelUpdate] Succeed: %s, Failed: %s',
        result['succeed_cnt'], result['failed_cnt'])

  def format_profit_rate(self, profit_rate):
    profit_level = math.floor(profit_rate / 10)
    return '{}-{}%'.format(profit_level * 10 + 1, (profit_level + 1) * 10)
