# -*- coding: utf-8 -*-

import datetime
import json
import time
import re
import multiprocessing
from multiprocessing.pool import ThreadPool
import pdb

import click
import dateparser
import shopify
import shopify_python_api_patcher
from shopify_python_api_patcher.utils import get_current_release

from em_tasks import logger
from em_tasks.shopify.product_util import ProductUtil
from em_tasks.store.store_util import StoreUtil
from em_tasks.utils.shopify_base_repricer import ShopifyBaseRepricer


class ShopifyNaverRepricer(ShopifyBaseRepricer):
  def __init__(self, store_code, product_service, price_calculator):
    super(ShopifyNaverRepricer, self).__init__()
    self.store = StoreUtil.get_store_by_code(store_code)
    if not self.store:
      return

    credential = json.loads(self.store.api_credential)
    self.product_util = ProductUtil(credential['name'], credential['token'], credential.get('release', None))

    self.product_service = product_service
    self.price_calc = price_calculator
    self.indice_name = 'naver_products'
    self.pool = ThreadPool(3)

  def run(self):
    batch_size = 250
    products = dict()
    for naver_prod in StoreUtil.get_naver_products():
      products[naver_prod['source_product_id']] = naver_prod

      if len(products) < batch_size:
        continue

      self.process_products(products)
      products.clear()

    if products:
      self.process_products(products)
      products.clear()

  def process_products(self, prods):
    product_ids = list(prods.keys())
    products = False
    max_retries = 3
    while max_retries > 0:
      try:
        products = self.product_service.search_products(self.indice_name, product_ids)
        if not isinstance(products, dict):
          logger.info('[ProductService] Temporary unavailable! Wait for 10 seconds to retry.')
          max_retries -= 1
          time.sleep(10)
          continue

        break
      except Exception as e:
        logger.exception(e)
        max_retries -= 1
        time.sleep(3)

    if not products:
      return

    for pid, product in products.items():
      if pid not in prods:
        continue

      prod = prods[pid]

      src_offer = None

      existence = product.get('existence', True)
      src_shipping_fee = 3
      currency = product.get('currency', None) or 'USD'
      if not existence:
        prod['offer'] = {'price': 0, 'quantity': 0}
      else:
        prod['offer'] = dict()
        variants = product.get('variants', [])
        if variants:
          for variant in variants:
            try:
              src_product_price = variant.get('price', 0)
              if not src_product_price:
                prod['offer'] = {'price': 0, 'quantity': 0}
              else:
                src_offer = {'price': src_product_price + src_shipping_fee, 'currency': currency}

                offer = self.price_calc.calc_offer(src_offer)
                prod['offer'][variant['sku']] = offer
            except:
              pdb.set_trace()
        else:
          src_product_price = product.get('price', 0)
          if not src_product_price:
            prod['offer'] = {'price': 0, 'quantity': 0}
          else:
            src_offer = {'price': src_product_price + src_shipping_fee, 'currency': currency}

            offer = self.price_calc.calc_offer(src_offer)
            prod['offer'] = offer

    offers = dict()
    for _, prod in prods.items():
      offers[prod['product_id']] = prod

    if offers:
      store_offers = self.product_util.set_products_offer(offers, self.pool)
      try:
        self.update_gmc_offers(store_offers)
      except Exception as e:
        logger.exception(e)
        handles = [store_offer['handle'] for store_offer in store_offers.values() if store_offer['handle']]
        logger.warning('[Handles] %s', handles)
