# -*- coding: utf-8 -*-

import pdb

from em_tasks import logger
from em_tasks.store.store_util import StoreUtil
from em_tasks.shopify.product_util import ProductUtil
from em_tasks.store.models import *
from em_tasks.utils.sku_parser import SkuParser


class ShopifyProductsSyncer():
  def __init__(self, store_code, next_page_url=None):
    self.store = StoreUtil.get_store_by_code(store_code)
    if not self.store:
      return

    credential = json.loads(self.store.api_credential)
    self.product_util = ProductUtil(credential['name'], credential['token'], credential.get('release', None))
    self.next_page_url = next_page_url

  def run(self):
    products_buf = dict()
    batch_size = 1000
    next_page_url = self.next_page_url
    max_retries = 12
    while max_retries > 0:
      try:
        for products, next_page_url in self.product_util.list_products(next_page_url):
          for product in products:
            products_buf[product.id] = product

          if len(products_buf) < batch_size:
            continue

          self.process_products(products_buf)
          products_buf.clear()

        break
      except Exception as e:
        logger.exception(e)
        max_retries -= 1

    if products_buf:
      self.process_products(products_buf)
      products_buf.clear()

  def process_products(self, products):
    to_create = []
    to_update = []
    product_ids = list(products.keys())
    for sp in StoreProduct.select().where(StoreProduct.s_product_id.in_(product_ids)).execute():
      if sp.s_product_id not in products:
        continue

      product = products.pop(sp.s_product_id)
      in_stock = False
      for v in product.variants:
        if v.inventory_quantity > 0:
          in_stock = True
          break

      variant = product.variants[0]
      record = {
        'id': sp.id,
        's_product_id': product.id,
        'handle': product.handle,
        'title': product.title,
        'description': product.body_html,
        'has_image': bool(product.images),
        'sku': variant.sku,
        'in_stock':in_stock
      }
      to_update.append(record)

    if to_update:
      StoreProduct.insert_many(to_update).on_conflict(
        preserve=(StoreProduct.handle, StoreProduct.title, StoreProduct.description, StoreProduct.has_image, StoreProduct.sku, StoreProduct.in_stock)).execute()
      to_update = []

    if not products:
      return

    for product_id, product in products.items():
      in_stock = False
      for v in product.variants:
        if v.inventory_quantity > 0:
          in_stock = True
          break

      variant = product.variants[0]
      record = {
        'store_seller_id': None,
        's_product_id': product.id,
        'handle': product.handle,
        'title': product.title,
        'description': product.body_html,
        'has_image': bool(product.images),
        'sku': variant.sku,
        'in_stock': in_stock
      }
      sid, source, source_product_id = SkuParser.parse(variant.sku)
      record['source'] = source
      record['source_product_id'] = source_product_id
      if sid:
        record['store_seller_id'] = sid
      else:
        if source == 'AMZ_US':
          record['store_seller_id'] = 1
        elif source == 'AMZ_JP':
          record['store_seller_id'] = 4
        elif source == 'AMZ_UK':
          record['store_seller_id'] = 8
        elif source == 'Coupang':
          record['store_seller_id'] = 5
        elif source == 'Rakuten':
          record['store_seller_id'] = 7
        elif source.startswith('Ebay'):
          record['store_seller_id'] = 1
        else:
          record['store_seller_id'] = None

      to_create.append(record)

    StoreProduct.insert_many(to_create).execute()
