# -*- coding: utf-8 -*-

import datetime
import json
import time
import re
import multiprocessing
from multiprocessing.pool import ThreadPool

import click
import dateparser

from em_tasks import logger
from em_tasks.spree.product_util import ProductUtil
from em_tasks.store.store_util import StoreUtil
from em_tasks.utils.spree.base_repricer import SpreeBaseRepricer
from em_tasks.utils.pipelines.ebay_products_filter import EbayProductsFilterPipeline


class SpreeEbayRepricer(SpreeBaseRepricer):
  def __init__(self, store_code, product_service, price_calculator, marketplace, ttl=14):
    super(SpreeEbayRepricer, self).__init__()
    self.store = StoreUtil.get_store_by_code(store_code)
    if not self.store:
      return

    credential = json.loads(self.store.api_credential)
    self.product_util = ProductUtil(credential['endpoint'], credential['api_key'], credential.get('api_version', 'v1'))
    # self.product_util = ProductUtil('http://localhost:3000', credential['api_key'], credential.get('api_version', 'v1'))

    self.product_service = product_service
    self.price_calc = price_calculator
    self.marketplace = marketplace
    self.indice_name = 'ebay_{}_products'.format(marketplace.lower())
    today = datetime.date.today()
    self.expire_date = today - datetime.timedelta(days=ttl)
    self.ebay_filter = EbayProductsFilterPipeline(20, 95)
    self.pool = ThreadPool(10)

  def run(self):
    batch_size = 250
    products = dict()
    for ebay_prod in self.product_util.get_ebay_products(self.marketplace):
      products[ebay_prod['source_product_id']] = ebay_prod

      if len(products) < batch_size:
        continue

      self.process_products(products)
      products.clear()

    if products:
      self.process_products(products)
      products = []

    self.pool.close()
    self.pool.join()

  def process_products(self, prods):
    product_ids = list(prods.keys())
    products = False
    max_retries = 3
    while max_retries > 0:
      try:
        products = self.product_service.search_products(self.indice_name, product_ids)
        if not isinstance(products, dict):
          logger.info('[ProductService] Temporary unavailable! Wait for 10 seconds to retry.')
          max_retries -= 1
          time.sleep(10)
          continue

        break
      except Exception as e:
        logger.exception(e)
        max_retries -= 1
        time.sleep(3)

    if not products:
      return

    for pid, product in products.items():
      if pid not in prods:
        continue

      prod = prods[pid]

      src_offer = None

      existence = product.get('existence', True)
      has_only_default_variant = product.get('has_only_default_variant', True)
      src_product_price = product.get('price', 0)
      src_shipping_fee = product.get('shipping_fee', 0) or 0
      currency = product.get('currency', None) or 'USD'
      result = self.ebay_filter.check_product(product)
      # if not existence or not has_only_default_variant or not src_product_price or (not result.get('passed', False)):
      if not existence or not src_product_price or (not result.get('passed', False)):
        if result.get('passed', False):
          logger.info('[ProductFiltered] %s', pid)

        prod['offer'] = {'price': 0, 'quantity': 0}
      else:
        update_date_s = product.get('date', product.get('time', None))
        if not update_date_s:
          prod['offer'] = {'price': 0, 'quantity': 0}
          continue

        update_date = None
        try:
          update_date = dateparser.parse(update_date_s).date()
        except:
          pass

        if not update_date:
          prod['offer'] = {'price': 0, 'quantity': 0}
          continue

        if update_date < self.expire_date:
          prod['offer'] = {'price': 0, 'quantity': 0}
          continue

        # shipping_days_max = product.get('shipping_days_max', None)
        # shipping_invalid = not shipping_days_max or shipping_days_max > 14
        shipping_invalid = False
        if not existence or not has_only_default_variant or not src_product_price or shipping_invalid:
          # Out of stock
          src_offer = {'price': src_product_price + src_shipping_fee, 'currency': currency}
        else:
          src_offer = {'price': src_product_price + src_shipping_fee, 'currency': currency}

        offer = self.price_calc.calc_offer(src_offer)
        if offer:
          prod['offer'] = dict(offer)
          prod['offer']['src_offer'] = src_offer
        else:
          prod['offer'] = offer

    offers = dict()
    for _, prod in prods.items():
      offers[prod['product_id']] = prod

    if offers:
      store_offers = self.product_util.set_products_offer(offers, self.pool)
      # try:
      #   self.update_gmc_offers(store_offers)
      # except Exception as e:
      #   logger.exception(e)
      #   handles = [store_offer['handle'] for store_offer in store_offers.values() if store_offer['handle']]
      #   logger.warning('[Handles] %s', handles)
