# -*- coding: utf-8 -*-

import datetime
import json
import time
import re
import multiprocessing
from multiprocessing.pool import ThreadPool

import click
import dateparser

from em_tasks import logger
from em_tasks.spree.product_util import ProductUtil
from em_tasks.store.store_util import StoreUtil
from em_tasks.utils.spree.base_repricer import SpreeBaseRepricer


class SpreeNaverRepricer(SpreeBaseRepricer):
  def __init__(self, store_code, product_service, price_calculator, ttl=7):
    super(SpreeNaverRepricer, self).__init__()
    self.store = StoreUtil.get_store_by_code(store_code)
    if not self.store:
      return

    credential = json.loads(self.store.api_credential)
    self.product_util = ProductUtil(credential['endpoint'], credential['api_key'], credential.get('api_version', 'v1'))
    # self.product_util = ProductUtil('http://localhost:3000', credential['api_key'], credential.get('api_version', 'v1'))

    self.product_service = product_service
    self.price_calc = price_calculator
    # self.indice_name = 'naver_products'
    self.indice_name = 'naver_elim'
    today = datetime.date.today()
    self.expire_date = today - datetime.timedelta(days=ttl)
    self.pool = ThreadPool(3)

  def run(self):
    batch_size = 250
    products = dict()
    for prod in self.product_util.get_naver_products():
      products[prod['source_product_id']] = prod

      if len(products) < batch_size:
        continue

      self.process_products(products)
      products.clear()

    if products:
      self.process_products(products)
      products = []

    self.pool.close()
    self.pool.join()

  def process_products(self, prods):
    product_ids = list(prods.keys())
    products = False
    max_retries = 3
    while max_retries > 0:
      try:
        products = self.product_service.search_products(self.indice_name, product_ids)
        if not isinstance(products, dict):
          logger.info('[ProductService] Temporary unavailable! Wait for 10 seconds to retry.')
          max_retries -= 1
          time.sleep(10)
          continue

        break
      except Exception as e:
        logger.exception(e)
        max_retries -= 1
        time.sleep(3)

    if not products:
      return

    for pid, product in products.items():
      if pid not in prods:
        continue

      prod = prods[pid]

      src_offer = None

      existence = product.get('existence', True)
      has_only_default_variant = product.get('has_only_default_variant', True)
      src_product_price = product.get('price', 0) or 0
      src_shipping_fee = product.get('shipping_fee', 0) or 0
      currency = product.get('currency', None) or 'USD'
      if not existence or not src_product_price:
        prod['offer'] = {'price': 0, 'quantity': 0}
      else:
        src_offer = {'price': src_product_price + src_shipping_fee, 'currency': currency}

        offer = self.price_calc.calc_offer(src_offer)
        if offer:
          prod['offer'] = dict(offer)
          prod['offer']['src_offer'] = src_offer
        else:
          prod['offer'] = offer

    offers = dict()
    for _, prod in prods.items():
      offers[prod['product_id']] = prod

    if offers:
      store_offers = self.product_util.set_products_offer(offers, self.pool)
