# -*- coding: utf-8 -*-

import datetime
import pdb

import dateutil
import dateparser

from tracking_tracker.ups_api import UPSApi

from em_tasks import logger
from em_tasks.store.store_util import StoreUtil
from em_tasks.store.models import *


class UPSTrackingTracker():
  def __init__(self, merchant_id, client_id, client_secret, environment='production', days=30):
    self.days = days
    self.merchant_id = merchant_id
    self.client_id = client_id
    self.client_secret = client_secret
    self.environment = environment
    self.api = UPSApi(merchant_id, client_id, client_secret, environment)

  def run(self):
    carrier = 'UPS'
    carrier_pattern = '%UPS%'
    now = datetime.datetime.now()
    from_date = now - datetime.timedelta(days=self.days)
    trackings = dict()
    for order_item in OrderItem.select(OrderItem.id, OrderItem.s_carrier, OrderItem.s_tracking).where((OrderItem.s_purchase_date >= from_date) & (OrderItem.s_carrier ** carrier_pattern)).execute():
      if not order_item.s_carrier or not order_item.s_tracking:
        continue

      trackings[order_item.s_tracking] = None

    for fulfill_order in FulfillOrder.select(FulfillOrder.id, FulfillOrder.buy_tracking).where(FulfillOrder.buy_carrier ** carrier_pattern).execute():
      if not fulfill_order.buy_tracking:
        continue

      trackings[fulfill_order.buy_tracking] = None

    for customer_return in CustomerReturn.select(CustomerReturn.id, CustomerReturn.tracking_number).where(CustomerReturn.carrier ** carrier_pattern).execute():
      if not customer_return.tracking_number:
        continue

      trackings[customer_return.tracking_number] = None

    tracking_numbers = list(trackings.keys())
    for tracking_timeline in TrackingTimeline.select().where(TrackingTimeline.tracking_number.in_(tracking_numbers)).execute():
      if tracking_timeline.delivered or tracking_timeline.not_exist:
        trackings.pop(tracking_timeline.tracking_number, None)
        continue

    if not trackings:
      return

    for tracking_number in trackings.keys():
      try:
        tracking = self.api.get_tracking_details(tracking_number)
        if not tracking or 'trackResponse' not in tracking:
          continue

        timelines = self.parse_timelines(tracking['trackResponse'])
        delivered = any([timeline['status'].lower().find('delivered') != -1 for timeline in timelines if timeline['status']])
        status = 'IN_TRANSIT'
        if delivered:
          status = 'DELIVERED'

        last_update_date = None
        if timelines:
          last_update = timelines[0]
          date_s = last_update['date']
          parsed_date = dateparser.parse(date_s)
          if parsed_date:
            last_update_date = parsed_date.strftime('%Y-%m-%d')
          else:
            last_update_date = date_s

        tracking_timeline = TrackingTimeline.get_or_none(TrackingTimeline.tracking_number == tracking_number)
        if tracking_timeline:
          tracking_timeline.delivered = delivered
          tracking_timeline.last_update_date =  last_update_date
          tracking_timeline.timelines = json.dumps(timelines)
          tracking_timeline.save()
        else:
          data = {
            'carrier': carrier,
            'tracking_number': tracking_number,
            'delivered': delivered,
            'last_update_date': last_update_date,
            'timelines': json.dumps(timelines)
          }
          TrackingTimeline.insert(data).execute()
          logger.info('[TrackingTimeline] %s', data)
      except Exception as e:
        logger.exception(e)

  def parse_timelines(self, tracking):
    if not tracking or 'shipment' not in tracking or not tracking['shipment']:
      return []

    timelines = []
    d = dict()
    shipment = tracking['shipment'][0]
    if 'package' not in shipment or not shipment['package']:
      return []
    activities = shipment['package'][0]['activity']
    for activity in activities:
      date_s = '{}T{}'.format(activity['date'], activity['time'])
      try:
        parsed_date = dateutil.parser.parse(date_s)
      except Exception as e:
        try:
          parsed_date = dateparser.parse(date_s)
        except Exception as e:
          # logger.exception(e)
          parsed_date = None
      if parsed_date:
        date_s = parsed_date.strftime('%m/%d/%Y')
      else:
        date_s = activity['date']

      address = activity['location'].get('address', d)
      if address:
        city = address.get('city', '')
        state = address.get('stateProvince', '')
        if state:
          location = '{}, {} {}'.format(
            city, state, address.get('postalCode', '')).strip()
        else:
          location = city.strip()
      else:
        location = ''

      state = activity['status']
      status = UPSApi.get_status_by_processing_code(state.get('statusCode', ''))
      if not status:
        if state['description'].lower().find('delivered') != -1:
          status = 'Delivered'
      timelines.append({
        'status': status or '',
        'status_detail': state['description'],
        'location': location,
        'date': date_s
      })

    return timelines
