# -*- coding: utf-8 -*-

import datetime
import pdb

import dateutil
import dateparser

from usps import USPSApi, USPSApiError

from em_tasks import logger
from em_tasks.store.store_util import StoreUtil
from em_tasks.store.models import *


class USPSTrackingTracker():
  def __init__(self, api_user_id, days=30):
    self.days = days
    self.api_user_id = api_user_id
    self.api = USPSApi(api_user_id)

  def run(self):
    carrier = 'USPS'
    carrier_pattern = '%USPS%'
    now = datetime.datetime.now()
    from_date = now - datetime.timedelta(days=self.days)
    trackings = dict()
    for order_item in OrderItem.select(OrderItem.id, OrderItem.s_carrier, OrderItem.s_tracking).where((OrderItem.s_purchase_date >= from_date) & (OrderItem.s_carrier ** carrier_pattern)).execute():
      if not order_item.s_carrier or not order_item.s_tracking:
        continue

      trackings[order_item.s_tracking] = None

    for fulfill_order in FulfillOrder.select(FulfillOrder.id, FulfillOrder.buy_tracking).where(FulfillOrder.buy_carrier ** carrier_pattern).execute():
      if not fulfill_order.buy_tracking:
        continue

      trackings[fulfill_order.buy_tracking] = None

    for customer_return in CustomerReturn.select(CustomerReturn.id, CustomerReturn.tracking_number).where(CustomerReturn.carrier ** carrier_pattern).execute():
      if not customer_return.tracking_number:
        continue

      trackings[customer_return.tracking_number] = None

    tracking_numbers = list(trackings.keys())
    for tracking_timeline in TrackingTimeline.select().where(TrackingTimeline.tracking_number.in_(tracking_numbers)).execute():
      if tracking_timeline.delivered or tracking_timeline.not_exist:
        trackings.pop(tracking_timeline.tracking_number, None)
        continue

    if not trackings:
      return

    d = dict()
    tracking_numbers = list(trackings.keys())
    tracking_timelines = dict()
    for tracking_timeline in TrackingTimeline.select().where(TrackingTimeline.tracking_number.in_(tracking_numbers)).execute():
      tracking_timelines[tracking_timeline.tracking_number] = tracking_timeline

    for tracking_number in trackings.keys():
      try:
        track = self.api.track(tracking_number)
        if not track or not track.result:
          logger.info('[InvalidResponse] %s', track)
          continue

        track_error = track.result.get('TrackResponse', d).get('TrackInfo', d).get('Error', None)
        if track_error:
          # tracking_timeline = TrackingTimeline.get_or_none(TrackingTimeline.tracking_number == tracking_number)
          # if not tracking_timeline:
          #   tracking_timeline = TrackingTimeline()
          #   tracking_timeline.carrier = carrier
          #   tracking_timeline.tracking_number = tracking_number

          # tracking_timeline.not_exist = True
          # tracking_timeline.save()
          continue

        timelines = self.parse_timelines(track)
        delivered = any([timeline['status'].lower().find('delivered') != -1 for timeline in timelines if timeline['status']])
        status = 'IN_TRANSIT'
        if delivered:
          status = 'DELIVERED'

        last_update_date = None
        if timelines:
          last_update = timelines[0]
          date_s = last_update['date']
          parsed_date = dateparser.parse(date_s)
          if parsed_date:
            last_update_date = parsed_date.strftime('%Y-%m-%d')
          else:
            last_update_date = date_s

        # tracking_timeline = TrackingTimeline.get_or_none(TrackingTimeline.tracking_number == tracking_number)
        # if tracking_timeline:
        if tracking_number in tracking_timelines:
          tracking_timeline = tracking_timelines[tracking_number]
          timelines_s = json.dumps(timelines)
          if tracking_timeline.delivered != delivered or tracking_timeline.timelines != timelines_s:
            logger.info('[TrackingUpdated] TrackingNumber: %s, Timelines: %s', tracking_number, timelines)

            tracking_timeline.delivered = delivered
            tracking_timeline.last_update_date =  last_update_date
            tracking_timeline.timelines = timelines_s
            tracking_timeline.save()
          else:
            logger.info(
              '[TrackingNoUpdated] TrackingNumber: %s, LastUpdateDate: %s',
              tracking_number, tracking_timeline.last_update_date)
        else:
          data = {
            'carrier': carrier,
            'tracking_number': tracking_number,
            'delivered': delivered,
            'last_update_date': last_update_date,
            'timelines': json.dumps(timelines)
          }
          TrackingTimeline.insert(data).execute()
          logger.info('[TrackingTimeline] %s', data)
      except USPSApiError as e:
        logger.exception(e)
      except Exception as e:
        logger.exception(e)

  def parse_timelines(self, track):
    d = dict()
    events = track.result.get('TrackResponse', d).get('TrackInfo', d).get('TrackDetail', [])
    summary = track.result.get('TrackResponse', d).get('TrackInfo', d).get('TrackSummary', d)
    if not events:
      if not summary:
        return []

      events = [summary]
    else:
      if isinstance(events, dict):
        events = [events]

      last_event = events[0]
      last_event_desc = last_event.get('Event', '')

      summary_desc = summary.get('Event', '')
      delivered = summary_desc.lower().find('delivered') != -1
      returned = summary_desc.lower().find('return') != -1
      if delivered:
        if last_event_desc.lower().find('delivered') == -1:
          events.insert(0, summary)

      if returned:
        if last_event_desc.lower().find('return') == -1:
          events.insert(0, summary)

    # if isinstance(events, dict):
    #   events = [events]

    timelines = []
    for event in events:
      try:
        date_s = event['EventDate']
        parsed_date = None
        try:
          parsed_date = dateutil.parser.parse(date_s)
        except:
          try:
            parsed_date = dateparser.parse(date_s)
          except Exception as e:
            # logger.exception(e)
            parsed_date = None
        if parsed_date:
          date_s = parsed_date.strftime('%m/%d/%Y')

        location_parts = [part for part in [event.get('EventCity', None), event.get('EventState', None), event.get('EventZIPCode', None)] if part]
        location = ','.join(location_parts)

        description = event.get('Event', '')
        status = 'IN_TRANSIT'
        if description.lower().find('delivered') != -1:
          status = 'Delivered'

        timelines.append({
          'status': status,
          'status_detail': description,
          'location': location,
          'date': date_s
        })
      except Exception as e:
        logger.exception(e)

    return timelines
